# Copyright (C) 2009 Canonical
#
# Authors:
#  Michael Vogt
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
from __future__ import print_function

import datetime
import os

from typing import Optional, Tuple

import apt
import apt_pkg


def get_maintenance_end_date(release_date, m_months):
    # type: (datetime.datetime, int) -> Tuple[int, int]
    """
    get the (year, month) tuple when the maintenance for the distribution
    ends. Needs the data of the release and the number of months that
    its is supported as input
    """
    # calc end date
    years = m_months // 12
    months = m_months % 12
    support_end_year = release_date.year + years + (release_date.month + months) // 12
    support_end_month = (release_date.month + months) % 12
    # special case: this happens when e.g. doing 2010-06 + 18 months
    if support_end_month == 0:
        support_end_month = 12
        support_end_year -= 1
    return (support_end_year, support_end_month)


def get_release_date_from_release_file(path):
    # type: (str) -> Optional[int]
    """
    return the release date as time_t for the given release file
    """
    if not path or not os.path.exists(path):
        return None

    with os.fdopen(apt_pkg.open_maybe_clear_signed_file(path)) as data:
        tag = apt_pkg.TagFile(data)
        section = next(tag)
        if "Date" not in section:
            return None
        date = section["Date"]
        return apt_pkg.str_to_time(date)


def get_release_filename_for_pkg(cache, pkgname, label, release):
    # type: (apt.Cache, str, str, str) -> Optional[str]
    "get the release file that provides this pkg"
    if pkgname not in cache:
        return None
    pkg = cache[pkgname]
    ver = None
    # look for the version that comes from the repos with
    # the given label and origin
    for aver in pkg._pkg.version_list:
        if aver is None or aver.file_list is None:
            continue
        for ver_file, _index in aver.file_list:
            # print verFile
            if (
                ver_file.origin == label
                and ver_file.label == label
                and ver_file.archive == release
            ):
                ver = aver
    if not ver:
        return None
    indexfile = cache._list.find_index(ver.file_list[0][0])
    for metaindex in cache._list.list:
        for m in metaindex.index_files:
            if indexfile and indexfile.describe == m.describe and indexfile.is_trusted:
                dirname = apt_pkg.config.find_dir("Dir::State::lists")
                for relfile in ["InRelease", "Release"]:
                    name = apt_pkg.uri_to_filename(metaindex.uri) + "dists_%s_%s" % (
                        metaindex.dist,
                        relfile,
                    )
                    if os.path.exists(dirname + name):
                        return dirname + name
    return None



import os
import unittest


if __name__ == "__main__":
    os.chdir(os.path.dirname(__file__))
    print(os.getcwd())

    for path in os.listdir("."):
        if path.endswith(".py"):
            exec("from %s import *" % path[:-3])
    unittest.main()



#!/usr/bin/python3

import unittest
import os
import copy
import tempfile

import apt_pkg
import aptsources.sourceslist
import aptsources.distro

import testcommon


class TestAptSources(testcommon.TestCase):
    def setUp(self):
        testcommon.TestCase.setUp(self)
        if apt_pkg.config["APT::Architecture"] not in ("i386", "amd64"):
            apt_pkg.config.set("APT::Architecture", "i386")
        apt_pkg.config.set("Dir::Etc", os.getcwd())
        apt_pkg.config.set("Dir::Etc::sourceparts", tempfile.mkdtemp())
        if os.path.exists("./build/data/templates"):
            self.templates = os.path.abspath("./build/data/templates")
        elif os.path.exists("../build/data/templates"):
            self.templates = os.path.abspath("../build/data/templates")
        else:
            self.templates = "/usr/share/python-apt/templates/"

    def tearDown(self):
        aptsources.distro._OSRelease.OS_RELEASE_FILE = (
            aptsources.distro._OSRelease.DEFAULT_OS_RELEASE_FILE
        )
        if "LSB_ETC_LSB_RELEASE" in os.environ:
            del os.environ["LSB_ETC_LSB_RELEASE"]

    def testIsMirror(self):
        """aptsources: Test mirror detection."""
        yes = aptsources.sourceslist.is_mirror(
            "http://archive.ubuntu.com", "http://de.archive.ubuntu.com"
        )
        no = aptsources.sourceslist.is_mirror(
            "http://archive.ubuntu.com", "http://ftp.debian.org"
        )
        self.assertTrue(yes)
        self.assertFalse(no)

    def testSourcesListReading(self):
        """aptsources: Test sources.list parsing."""
        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        sources = aptsources.sourceslist.SourcesList(True, self.templates)
        self.assertEqual(len(sources.list), 10)
        # test load
        sources.list = []
        sources.load("data/aptsources/sources.list")
        self.assertEqual(len(sources.list), 10)

    def testSourcesListReading_deb822(self):
        """aptsources: Test sources.list parsing."""
        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        self.assertEqual(len(sources.list), 5)
        # test load
        sources.list = []
        sources.load("data/aptsources/sources.list.d/main.sources")
        self.assertEqual(len(sources.list), 5)

        for entry in sources.list:
            self.assertFalse(entry.invalid)
            self.assertFalse(entry.disabled)
            self.assertEqual(entry.types, ["deb"])
            self.assertEqual(entry.type, "deb")
            self.assertEqual(entry.uris, ["http://de.archive.ubuntu.com/ubuntu/"])
            self.assertEqual(entry.uri, "http://de.archive.ubuntu.com/ubuntu/")

        self.assertEqual(sources.list[0].comps, ["main"])
        self.assertEqual(sources.list[1].comps, ["restricted"])
        self.assertEqual(sources.list[2].comps, ["universe"])
        self.assertEqual(sources.list[3].comps, ["main"])
        self.assertEqual(sources.list[4].comps, ["main"])

        for entry in sources.list[:-1]:
            self.assertIsNone(entry.trusted)

        self.assertTrue(sources.list[-1].trusted)

        for entry in sources.list[:-2]:
            self.assertEqual(entry.architectures, [])
            self.assertEqual(entry.suites, ["edgy"])
            self.assertEqual(entry.dist, "edgy")

        for entry in sources.list[-2:]:
            self.assertEqual(entry.suites, ["natty"])
            self.assertEqual(entry.dist, "natty")
            self.assertEqual(entry.architectures, ["amd64", "i386"])

    def testSourcesListWriting_deb822(self):
        """aptsources: Test sources.list parsing."""
        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        with tempfile.TemporaryDirectory() as tmpdir:
            for entry in sources.list:
                entry.file = os.path.join(tmpdir, os.path.basename(entry.file))

            sources.save()

            maxDiff = self.maxDiff
            self.maxDiff = None
            for file in os.listdir("data/aptsources/sources.list.d"):
                with open(os.path.join("data/aptsources/sources.list.d", file)) as a:
                    with open(os.path.join(tmpdir, file)) as b:
                        self.assertEqual(a.read(), b.read(), f"file {file}")
            self.maxDiff = maxDiff

    def testSourcesListAdding(self):
        """aptsources: Test additions to sources.list"""
        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        sources = aptsources.sourceslist.SourcesList(True, self.templates)
        # test to add something that is already there (main)
        before = copy.deepcopy(sources)
        sources.add("deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["main"])
        self.assertTrue(sources.list == before.list)
        # test to add something that is already there (restricted)
        before = copy.deepcopy(sources)
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["restricted"]
        )
        self.assertTrue(sources.list == before.list)

        before = copy.deepcopy(sources)
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "natty",
            ["main"],
            architectures=["amd64", "i386"],
        )
        self.assertTrue(sources.list == before.list)

        # test to add something new: multiverse
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["multiverse"]
        )
        found = False
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "edgy"
                and "multiverse" in entry.comps
            ):
                found = True
                break
        self.assertTrue(found)

        # add a new natty entry without architecture specification
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "natty", ["multiverse"]
        )
        found = False
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "natty"
                and entry.architectures == []
                and "multiverse" in entry.comps
            ):
                found = True
                break
        self.assertTrue(found)

        # Add universe to existing multi-arch line
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "natty",
            ["universe"],
            architectures=["i386", "amd64"],
        )
        found = False
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "natty"
                and set(entry.architectures) == set(["amd64", "i386"])
                and set(entry.comps) == set(["main", "universe"])
            ):
                found = True
                break
        self.assertTrue(found)
        # test to add something new: multiverse *and*
        # something that is already there
        before = copy.deepcopy(sources)
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "edgy",
            ["universe", "something"],
        )
        found_universe = 0
        found_something = 0
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "edgy"
            ):
                for c in entry.comps:
                    if c == "universe":
                        found_universe += 1
                    if c == "something":
                        found_something += 1
        # print "\n".join([s.str() for s in sources])
        self.assertEqual(found_something, 1)
        self.assertEqual(found_universe, 1)

    def testSourcesListAdding_deb822(self):
        """aptsources: Test additions to sources.list"""
        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")

        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        # test to add something that is already there (main)
        before = copy.deepcopy(sources)
        sources.add("deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["main"])
        self.assertEqual(sources.list, before.list)
        # test to add something that is already there (restricted)
        before = copy.deepcopy(sources)
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["restricted"]
        )
        self.assertTrue(sources.list == before.list)

        before = copy.deepcopy(sources)
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "natty",
            ["main"],
            architectures=["amd64", "i386"],
        )
        self.assertEqual(
            [e.str() for e in sources.list], [e.str() for e in before.list]
        )

        # test to add something new: multiverse
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["multiverse"]
        )
        found = False
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "edgy"
                and "multiverse" in entry.comps
            ):
                found = True
                break
        self.assertTrue(found)

        # add a new natty entry without architecture specification
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "natty", ["multiverse"]
        )
        found = False
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "natty"
                and entry.architectures == []
                and "multiverse" in entry.comps
            ):
                found = True
                break
        self.assertTrue(found)

        # Add universe to existing multi-arch line
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "natty",
            ["universe"],
            architectures=["i386", "amd64"],
        )
        found = False
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "natty"
                and set(entry.architectures) == set(["amd64", "i386"])
                and set(entry.comps) == set(["main", "universe"])
            ):
                found = True
                break
        self.assertTrue(found)
        # test to add something new: multiverse *and*
        # something that is already there
        before = copy.deepcopy(sources)
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "edgy",
            ["universe", "something"],
        )
        found_universe = 0
        found_something = 0
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "edgy"
            ):
                for c in entry.comps:
                    if c == "universe":
                        found_universe += 1
                    if c == "something":
                        found_something += 1
        # print "\n".join([s.str() for s in sources])
        self.assertEqual(found_something, 1)
        self.assertEqual(found_universe, 1)

    def testAddingWithComment_deb822(self):
        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        self._commonTestAddingWithComment()

    def testAddingWithComment_short(self):
        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        self._commonTestAddingWithComment()

    def _commonTestAddingWithComment(self):
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)

        # test to add something that is already there (main); loses comment
        before = copy.deepcopy(sources)
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "edgy",
            ["main"],
            comment="this will be lost",
        )
        self.assertTrue(sources.list == before.list)
        for entry in sources:
            self.assertNotEqual(entry.comment, "this will be lost")

        # test to add component to existing entry: multiverse; loses comment
        sources.add(
            "deb",
            "http://de.archive.ubuntu.com/ubuntu/",
            "edgy",
            ["multiverse"],
            comment="this will be lost",
        )
        for entry in sources:
            self.assertNotEqual(entry.comment, "this will be lost")

        before = copy.deepcopy(sources)
        # test to add entirely new entry; retains comment
        sources.add(
            "deb-src",
            "http://de.archive.ubuntu.com/ubuntu/",
            "edgy",
            ["main"],
            comment="this will appear",
            file=sources.list[0].file,  # make sure we test the deb822 code path
        )
        self.assertNotEqual(sources.list, before.list)
        self.assertEqual(len(sources.list), len(before.list) + 1)
        found = False
        for entry in sources:
            if (
                entry.type == "deb-src"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and entry.dist == "edgy"
                and entry.comment == "this will appear"
                and "main" in entry.comps
            ):
                found = True
                break
        self.assertTrue(found)

    def testInsertion_deb822(self):
        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)

        # test to insert something that is already there (universe); does not
        # move existing entry (remains at index 2)
        before = copy.deepcopy(sources)
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["universe"], pos=0
        )
        self.assertTrue(sources.list == before.list)
        entry = list(sources)[2]
        self.assertEqual(entry.type, "deb")
        self.assertEqual(entry.uri, "http://de.archive.ubuntu.com/ubuntu/")
        self.assertEqual(entry.dist, "edgy")
        self.assertIn("universe", entry.comps)

        # test add component to existing entry: multiverse; does not move
        # entry to which it is appended (remains at index 0)
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["multiverse"], pos=2
        )
        entry = list(sources)[0]
        self.assertTrue(
            entry.type == "deb"
            and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
            and entry.dist == "edgy"
            and {"main", "multiverse"} <= set(entry.comps)
        )

        # test to add entirely new entry; inserted at 0
        sources.add(
            "deb-src", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["main"], pos=0
        )
        entry = list(sources)[0]
        self.assertTrue(
            entry.type == "deb-src"
            and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
            and entry.dist == "edgy"
            and "main" in entry.comps
        )

    def testInsertion(self):
        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        sources = aptsources.sourceslist.SourcesList(True, self.templates)

        # test to insert something that is already there (universe); does not
        # move existing entry (remains at index 2)
        before = copy.deepcopy(sources)
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["universe"], pos=0
        )
        self.assertTrue(sources.list == before.list)
        entry = list(sources)[5]
        self.assertTrue(
            entry.type == "deb"
            and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
            and entry.dist == "edgy"
            and "universe" in entry.comps
        )

        # test add component to existing entry: multiverse; does not move
        # entry to which it is appended (remains at index 0)
        sources.add(
            "deb", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["multiverse"], pos=2
        )
        entry = list(sources)[1]
        self.assertTrue(
            entry.type == "deb"
            and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
            and entry.dist == "edgy"
            and {"main", "multiverse"} <= set(entry.comps)
        )

        # test to add entirely new entry; inserted at 0
        sources.add(
            "deb-src", "http://de.archive.ubuntu.com/ubuntu/", "edgy", ["main"], pos=0
        )
        entry = list(sources)[0]
        self.assertTrue(
            entry.type == "deb-src"
            and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
            and entry.dist == "edgy"
            and "main" in entry.comps
        )

    def testDuplication_short(self):
        apt_pkg.config.set(
            "Dir::Etc::sourcelist", "data/aptsources/sources.list.testDuplication"
        )
        return self.commonTestDuplication()

    def testDuplication_deb822(self):
        apt_pkg.config.set(
            "Dir::Etc::sourceparts", "data/aptsources/sources.list.d.testDuplication"
        )
        return self.commonTestDuplication()

    def commonTestDuplication(self):
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        test_url = "http://ppa.launchpad.net/me/myproject/ubuntu"
        # test to add something that is already there (enabled)
        before = copy.deepcopy(sources)
        sources.add("deb", test_url, "xenial", ["main"])
        self.assertEqual(sources.list, before.list)
        # test to add something that is already there (disabled)
        sources.add("# deb-src", test_url, "xenial", ["main"])
        self.assertEqual(sources.list, before.list)
        # test to enable something that is already there
        sources.add("deb-src", test_url, "xenial", ["main"])
        found = False
        self.assertEqual(len(sources.list), 2)
        for entry in sources:
            if (
                entry.type == "deb-src"
                and not entry.disabled
                and entry.uri == test_url
                and entry.dist == "xenial"
                and entry.architectures == []
                and entry.comps == ["main"]
            ):
                found = True
                break
        self.assertTrue(found)

    def testMatcher_short(self):
        """aptsources: Test matcher"""
        apt_pkg.config.set(
            "Dir::Etc::sourcelist", "data/aptsources/" "sources.list.testDistribution"
        )
        return self.commonTestMatcher()

    def testMatcher_deb822(self):
        """aptsources: Test matcher"""
        apt_pkg.config.set(
            "Dir::Etc::sourceparts",
            "data/aptsources/" "sources.list.d.testDistribution",
        )
        return self.commonTestMatcher()

    def commonTestMatcher(self):
        """aptsources: Test matcher"""
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        distro = aptsources.distro.get_distro(
            id="Ubuntu",
            codename="bionic",
            description="Ubuntu 18.04 LTS",
            release="18.04",
        )
        distro.get_sources(sources)
        # test if all suits of the current distro were detected correctly
        self.assertNotEqual(sources.list, [])
        for s in sources:
            if not s.template:
                self.fail("source entry '%s' has no matcher" % s)

        # Hack in a check for splitting of fields here.
        if sources.list[-1].file.endswith(".sources"):
            self.assertEqual(
                sources.list[-1].uris, ["cdrom:[Ubuntu 8.04 _Hardy Heron_]"]
            )
        self.assertEqual(sources.list[-1].uri, "cdrom:[Ubuntu 8.04 _Hardy Heron_]")

    def testMultiArch(self):
        """aptsources: Test multi-arch parsing"""

        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        assert not sources.list[8].invalid
        assert sources.list[8].type == "deb"
        assert sources.list[8].architectures == ["amd64", "i386"]
        assert sources.list[8].uri == "http://de.archive.ubuntu.com/ubuntu/"
        assert sources.list[8].dist == "natty"
        assert sources.list[8].comps == ["main"]
        assert sources.list[8].line.strip() == str(sources.list[8])
        assert sources.list[8].trusted is None

    def testMultiArch_deb822(self):
        """aptsources: Test multi-arch parsing"""

        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        assert not sources.list[3].invalid
        assert sources.list[3].type == "deb"
        assert sources.list[3].architectures == ["amd64", "i386"]
        assert sources.list[3].uri == "http://de.archive.ubuntu.com/ubuntu/"
        assert sources.list[3].dist == "natty"
        assert sources.list[3].comps == ["main"]
        self.assertEqual(sources.list[3].line.strip(), str(sources.list[3]))
        self.assertIsNone(sources.list[3].trusted)

    def testMultipleOptions(self):
        """aptsources: Test multi-arch parsing"""

        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        sources = aptsources.sourceslist.SourcesList(True, self.templates)
        assert sources.list[9].invalid is False
        assert sources.list[9].type == "deb"
        assert sources.list[9].architectures == ["amd64", "i386"]
        self.assertEqual(sources.list[9].uri, "http://de.archive.ubuntu.com/ubuntu/")
        assert sources.list[9].dist == "natty"
        assert sources.list[9].comps == ["main"]
        assert sources.list[9].trusted
        assert sources.list[9].line.strip() == str(sources.list[9])

    def testMultipleOptions_deb822(self):
        """aptsources: Test multi-arch parsing"""

        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        assert sources.list[4].invalid is False
        assert sources.list[4].type == "deb"
        assert sources.list[4].architectures == ["amd64", "i386"]
        self.assertEqual(sources.list[4].uri, "http://de.archive.ubuntu.com/ubuntu/")
        assert sources.list[4].dist == "natty"
        assert sources.list[4].comps == ["main"]
        assert sources.list[4].trusted
        assert sources.list[4].line.strip() == str(sources.list[4])

    def test_enable_component(self):
        target = "./data/aptsources/sources.list.enable_comps"
        line = "deb http://archive.ubuntu.com/ubuntu lucid main\n"
        with open(target, "w") as target_file:
            target_file.write(line)
        apt_pkg.config.set("Dir::Etc::sourcelist", target)
        sources = aptsources.sourceslist.SourcesList(True, self.templates)
        distro = aptsources.distro.get_distro(id="Ubuntu")
        # make sure we are using the right distro
        distro.codename = "lucid"
        distro.id = "Ubuntu"
        distro.release = "10.04"
        # and get the sources
        distro.get_sources(sources)
        # test enable_component
        comp = "multiverse"
        distro.enable_component(comp)
        comps = set()
        for entry in sources:
            comps = comps.union(set(entry.comps))
        self.assertTrue("multiverse" in comps)
        self.assertTrue("universe" in comps)

    def test_enable_component_deb822(self):
        target = (
            apt_pkg.config.find_dir("dir::etc::sourceparts") + "enable_comps.sources"
        )
        line = "Types: deb\nURIs: http://archive.ubuntu.com/ubuntu\nSuites: lucid\nComponents: main\n"
        with open(target, "w") as target_file:
            target_file.write(line)
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        distro = aptsources.distro.get_distro(id="Ubuntu")
        # make sure we are using the right distro
        distro.codename = "lucid"
        distro.id = "Ubuntu"
        distro.release = "10.04"
        # and get the sources
        distro.get_sources(sources)
        # test enable_component
        comp = "multiverse"
        distro.enable_component(comp)
        comps = set()
        for entry in sources:
            comps = comps.union(set(entry.comps))
        self.assertTrue("multiverse" in comps)
        self.assertTrue("universe" in comps)

    def testDistribution_short(self):
        """aptsources: Test distribution detection."""
        apt_pkg.config.set(
            "Dir::Etc::sourcelist", "data/aptsources/" "sources.list.testDistribution"
        )
        return self.commonTestDistribution()

    def testDistribution_deb822(self):
        """aptsources: Test distribution detection."""
        apt_pkg.config.set(
            "Dir::Etc::sourceparts",
            "data/aptsources/" "sources.list.d.testDistribution",
        )
        return self.commonTestDistribution()

    def commonTestDistribution(self):
        """aptsources: Test distribution detection."""
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        distro = aptsources.distro.get_distro(
            id="Ubuntu",
            codename="bionic",
            description="Ubuntu 18.04 LTS",
            release="18.04",
        )
        distro.get_sources(sources)
        # test if all suits of the current distro were detected correctly
        dist_templates = set()
        for s in sources:
            if s.template:
                dist_templates.add(s.template.name)
        # print dist_templates
        for d in (
            "hardy",
            "hardy-security",
            "hardy-updates",
            "intrepid",
            "hardy-backports",
        ):
            self.assertTrue(d in dist_templates)
        # test enable
        comp = "restricted"
        distro.enable_component(comp)
        found = {}
        for entry in sources:
            if (
                entry.type == "deb"
                and entry.uri == "http://de.archive.ubuntu.com/ubuntu/"
                and "edgy" in entry.dist
            ):
                for c in entry.comps:
                    if c == comp:
                        if entry.dist not in found:
                            found[entry.dist] = 0
                        found[entry.dist] += 1
        # print "".join([s.str() for s in sources])
        for key in found:
            self.assertEqual(found[key], 1)

        # add a not-already available component
        comp = "multiverse"
        distro.enable_component(comp)
        found = {}
        for entry in sources:
            if entry.type == "deb" and entry.template and entry.template.name == "edgy":
                for c in entry.comps:
                    if c == comp:
                        if entry.dist not in found.has_key:
                            found[entry.dist] = 0
                        found[entry.dist] += 1
        # print "".join([s.str() for s in sources])
        for key in found:
            self.assertEqual(found[key], 1)

    @unittest.skip("lsb-release test broken when it was added")
    def test_os_release_distribution(self):
        """os-release file can be read and is_like is populated accordingly"""
        os.environ["LSB_ETC_LSB_RELEASE"] = os.path.abspath(
            "./data/aptsources/lsb-release"
        )
        aptsources.distro._OSRelease.OS_RELEASE_FILE = os.path.abspath(
            "./data/aptsources/os-release"
        )
        distro = aptsources.distro.get_distro()
        # Everything but is_like comes from lsb_release, see TODO in
        # get_distro.
        self.assertEqual("Ubuntu", distro.id)
        self.assertEqual("xenial", distro.codename)
        self.assertEqual("Ubuntu 16.04.1 LTS", distro.description)
        self.assertEqual("16.04", distro.release)
        self.assertEqual(["ubuntu", "debian"], distro.is_like)

    def test_enable_disabled_short(self):
        """LP: #1042916: Test enabling disabled entry."""
        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        return self.common_test_enable_disabled()

    def test_enable_disabled_deb822(self):
        """LP: #1042916: Test enabling disabled entry."""
        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        return self.common_test_enable_disabled()

    def common_test_enable_disabled(self):
        """LP: #1042916: Test enabling disabled entry."""
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        disabled = sources.add(
            "deb",
            "http://fi.archive.ubuntu.com/ubuntu/",
            "precise",
            ["main"],
            file=sources.list[0].file,  # if we use deb822, enable deb822
        )
        disabled.set_enabled(False)
        enabled = sources.add(
            "deb",
            "http://fi.archive.ubuntu.com/ubuntu/",
            "precise",
            ["main"],
            file=sources.list[0].file,  # if we use deb822, enable deb822
        )
        self.assertEqual(disabled, enabled)
        self.assertFalse(disabled.disabled)

    def test_duplicate_uri_with_trailing_slash_short(self):
        """Test replacing entry with same uri except trailing slash"""
        apt_pkg.config.set("Dir::Etc::sourcelist", "data/aptsources/" "sources.list")
        return self.common_test_duplicate_uri_with_trailing_slash()

    def test_duplicate_uri_with_trailing_slash_deb822(self):
        """Test replacing entry with same uri except trailing slash"""
        apt_pkg.config.set("Dir::Etc::sourceparts", "data/aptsources/" "sources.list.d")
        return self.common_test_duplicate_uri_with_trailing_slash()

    def common_test_duplicate_uri_with_trailing_slash(self):
        """Test replacing entry with same uri except trailing slash"""
        sources = aptsources.sourceslist.SourcesList(True, self.templates, deb822=True)
        line_wslash = "deb http://rslash.ubuntu.com/ubuntu/ precise main"
        line_woslash = "deb http://rslash.ubuntu.com/ubuntu precise main"
        entry_wslash = aptsources.sourceslist.SourceEntry(line_wslash)
        entry_woslash = aptsources.sourceslist.SourceEntry(line_woslash)
        self.assertEqual(entry_wslash, entry_woslash)
        count = len(sources.list)
        sourceslist_wslash = sources.add(
            entry_wslash.type, entry_wslash.uri, entry_wslash.dist, entry_wslash.comps
        )
        self.assertEqual(count + 1, len(sources.list))
        count = len(sources.list)
        sourceslist_woslash = sources.add(
            entry_woslash.type,
            entry_woslash.uri,
            entry_woslash.dist,
            entry_woslash.comps,
        )
        self.assertEqual(count, len(sources.list))
        self.assertEqual(sourceslist_wslash, sourceslist_woslash)


if __name__ == "__main__":
    os.chdir(os.path.dirname(__file__))
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.
"""Unit tests for verifying the correctness of DebFile descriptor handling."""
import os
import unittest

from test_all import get_library_dir
import sys

libdir = get_library_dir()
if libdir:
    sys.path.insert(0, libdir)
import apt_inst
import gc
import subprocess
import tempfile
import warnings


@unittest.skipIf(not os.path.exists("/proc/self/fd"), "no /proc/self/fd available")
class TestCVE_2020_27351(unittest.TestCase):
    """test the debfile"""

    GOOD_DEB = "data/test_debs/utf8-package_1.0-1_all.deb"

    def test_success(self):
        """opening package successfully should not leak fd"""
        before = os.listdir("/proc/self/fd")
        apt_inst.DebFile(self.GOOD_DEB)
        after = os.listdir("/proc/self/fd")
        self.assertEqual(before, after)

    def test_regression_bug_977000(self):
        """opening with a file handle should work correctly"""
        with open(self.GOOD_DEB) as good_deb:
            apt_inst.DebFile(good_deb).control.extractdata("control")

    def test_regression_bug_977000_2(self):
        """file object <-> debfile cycles should be collected by gc."""

        class Cycle(object):
            def __init__(self, fname):
                self.file = open(fname)
                self.deb = apt_inst.DebFile(self)

            def fileno(self):
                return self.file.fileno()

        before = os.listdir("/proc/self/fd")
        Cycle(self.GOOD_DEB).deb.control.extractdata("control")
        warnings.filterwarnings("ignore", category=ResourceWarning)
        gc.collect()
        warnings.resetwarnings()
        after = os.listdir("/proc/self/fd")
        self.assertEqual(before, after)

    def test_regression_bug_977000_2_ar(self):
        """file object <-> debfile cycles should be collected by gc."""

        class Cycle(object):
            def __init__(self, fname):
                self.file = open(fname)
                self.deb = apt_inst.ArArchive(self)

            def fileno(self):
                return self.file.fileno()

        before = os.listdir("/proc/self/fd")
        Cycle(self.GOOD_DEB).deb.gettar("control.tar.gz", "gzip").extractdata("control")
        warnings.filterwarnings("ignore", category=ResourceWarning)
        gc.collect()
        warnings.resetwarnings()
        after = os.listdir("/proc/self/fd")
        self.assertEqual(before, after)

    def test_success_a_member(self):
        """fd should be kept around as long as a tarfile member"""
        before = os.listdir("/proc/self/fd")
        data = apt_inst.DebFile(self.GOOD_DEB).data
        after = os.listdir("/proc/self/fd")
        self.assertEqual(len(before), len(after) - 1)
        del data
        after = os.listdir("/proc/self/fd")
        self.assertEqual(before, after)

    def _create_deb_without(self, member):
        temp = tempfile.NamedTemporaryFile(mode="wb")
        try:
            with open(self.GOOD_DEB, "rb") as deb:
                temp.write(deb.read())
            temp.flush()
            subprocess.check_call(["ar", "d", temp.name, member])
            return temp
        except Exception as e:
            temp.close()
            raise e

    def test_nocontrol(self):
        """opening package without control.tar.gz should not leak fd"""
        before = os.listdir("/proc/self/fd")
        with self._create_deb_without("control.tar.gz") as temp:
            try:
                apt_inst.DebFile(temp.name)
            except SystemError as e:
                self.assertIn("control.tar", str(e))
            else:
                self.fail("Did not raise an exception")

        after = os.listdir("/proc/self/fd")
        self.assertEqual(before, after)

    def test_nodata(self):
        """opening package without data.tar.gz should not leak fd"""
        before = os.listdir("/proc/self/fd")
        with self._create_deb_without("data.tar.gz") as temp:
            try:
                apt_inst.DebFile(temp.name)
            except SystemError as e:
                self.assertIn("data.tar", str(e))
            else:
                self.fail("Did not raise an exception")

        after = os.listdir("/proc/self/fd")
        self.assertEqual(before, after)

    def test_no_debian_binary(self):
        """opening package without debian-binary should not leak fd"""
        before = os.listdir("/proc/self/fd")
        with self._create_deb_without("debian-binary") as temp:
            try:
                apt_inst.DebFile(temp.name)
            except SystemError as e:
                self.assertIn("missing debian-binary", str(e))
            else:
                self.fail("Did not raise an exception")

        after = os.listdir("/proc/self/fd")
        self.assertEqual(before, after)


if __name__ == "__main__":
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



#!/usr/bin/python3
# this is a example how to access the build dependencies of a package

import apt
import apt_pkg
import sys

# main
cache = apt.Cache()
srcrecords = apt_pkg.SourceRecords()

# base package that we use for build-depends calculation
if len(sys.argv) < 2:
    print("need a package name as argument")
    sys.exit(1)
try:
    pkg = base = cache[sys.argv[1]]
except KeyError:
    print("No package %s found" % sys.argv[1])
    sys.exit(1)
all_build_depends = set()

# get the build depdends for the package itself
srcpkg_name = base.candidate.source_name
print("srcpkg_name: %s " % srcpkg_name)
if not srcpkg_name:
    print("Can't find source package for '%s'" % pkg.name)
srcrec = srcrecords.lookup(srcpkg_name)
if srcrec:
    print("Files:")
    print(srcrecords.files)
    bd = srcrecords.build_depends
    print("build-depends of the package: %s " % bd)
    for b in bd:
        all_build_depends.add(b[0])

# calculate the build depends for all dependencies
depends = base.candidate.dependencies
for or_dep in depends:
    for dep in or_dep.or_dependencies:
        pkg = cache[dep.name]
        srcpkg_name = pkg.candidate.source_name
        if not srcpkg_name:
            print("Can't find source package for '%s'" % pkg.name)
            continue
        srcrec = srcrecords.lookup(srcpkg_name)
        if srcrec:
            # print srcrecords.package
            # print srcrecords.binaries
            bd = srcrecords.build_depends
            # print "%s: %s " % (srcpkg_name, bd)
            for b in bd:
                all_build_depends.add(b[0])


print("\n".join(all_build_depends))



#!/usr/bin/python3
import os
import unittest
import tempfile

import apt_pkg
import aptsources.sourceslist
import aptsources.distro

import testcommon


class TestAptSourcesPorts(testcommon.TestCase):
    """Test aptsources on ports.ubuntu.com."""

    def setUp(self):
        testcommon.TestCase.setUp(self)
        apt_pkg.config.set("APT::Architecture", "powerpc")
        apt_pkg.config.set("Dir::Etc", os.path.abspath("data/aptsources_ports"))
        apt_pkg.config.set("Dir::Etc::sourceparts", tempfile.mkdtemp())
        if os.path.exists("../build/data/templates"):
            self.templates = os.path.abspath("../build/data/templates")
        else:
            self.templates = "/usr/share/python-apt/templates/"

    def testMatcher(self):
        """aptsources_ports: Test matcher."""
        apt_pkg.config.set("Dir::Etc::sourcelist", "sources.list")
        sources = aptsources.sourceslist.SourcesList(True, self.templates)
        distro = aptsources.distro.get_distro("Ubuntu", "hardy", "desc", "8.04")
        distro.get_sources(sources)
        # test if all suits of the current distro were detected correctly
        for s in sources:
            if not s.line.strip() or s.line.startswith("#"):
                continue
            if not s.template:
                self.fail("source entry '%s' has no matcher" % s)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# Copyright (C) 2009 Julian Andres Klode <jak@debian.org>
#
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.
"""Run all available unit tests."""
import os
import unittest.runner
import unittest
import sys

# Python 3 only provides abiflags since 3.2
if not hasattr(sys, "pydebug"):
    if sys.abiflags.startswith("d"):
        sys.pydebug = True
    else:
        sys.pydebug = False


def get_library_dir():
    # Find the path to the built apt_pkg and apt_inst extensions
    if not os.path.exists("../build"):
        return None
    from sysconfig import get_platform, get_python_version

    # Set the path to the build directory.
    plat_specifier = ".%s-%s" % (get_platform(), get_python_version())
    library_dir = "../build/lib%s%s" % (
        plat_specifier,
        (sys.pydebug and "-pydebug" or ""),
    )
    return os.path.abspath(library_dir)


class MyTestRunner(unittest.runner.TextTestRunner):
    def __init__(self, *args, **kwargs):
        kwargs["stream"] = sys.stdout
        super(MyTestRunner, self).__init__(*args, **kwargs)


if __name__ == "__main__":
    if not os.access("/etc/apt/sources.list", os.R_OK):
        sys.stdout.write("[tests] Skipping because sources.list is not readable\n")
        sys.exit(0)

    sys.stdout.write("[tests] Running on %s\n" % sys.version.replace("\n", ""))
    dirname = os.path.dirname(__file__)
    if dirname:
        os.chdir(dirname)
    library_dir = get_library_dir()
    if "pybuild" in os.getenv("PYTHONPATH", ""):
        # pybuild already supplied us with a path to check for
        sys.stdout.write("Using pybuild supplied build dir\n")
    elif library_dir:
        sys.stdout.write("Using library_dir: '%s'\n" % library_dir)
        sys.path.insert(0, os.path.abspath(library_dir))

    for path in os.listdir("."):
        if path.endswith(".py") and os.path.isfile(path) and path.startswith("test_"):
            exec("from %s import *" % path[:-3])

    unittest.main(testRunner=MyTestRunner)



#!/usr/bin/python3
#
# Copyright (C) 2010 Michael Vogt <michael.vogt@ubuntu.com>
#
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.

import datetime
import os
import unittest

from apt.utils import (
    get_maintenance_end_date,
    get_release_date_from_release_file,
)

import testcommon


class TestUtils(testcommon.TestCase):
    def test_get_release_date_from_release_file(self):
        path = os.path.join(os.path.dirname(__file__), "data", "misc", "foo_Release")
        t = get_release_date_from_release_file(path)
        self.assertEqual(
            str(datetime.datetime.utcfromtimestamp(t)), "2012-04-25 22:49:23"
        )

    def test_maintenance_time(self):
        months_of_support = 18
        # test historic releases, jaunty
        release_date = datetime.datetime(2009, 4, 23)
        (end_year, end_month) = get_maintenance_end_date(
            release_date, months_of_support
        )
        self.assertEqual(end_year, 2010)
        self.assertEqual(end_month, 10)
        # test historic releases, karmic
        release_date = datetime.datetime(2009, 10, 29)
        (end_year, end_month) = get_maintenance_end_date(
            release_date, months_of_support
        )
        self.assertEqual(end_year, 2011)
        self.assertEqual(end_month, 4)
        # test maverick
        release_date = datetime.datetime(2010, 10, 10)
        (end_year, end_month) = get_maintenance_end_date(
            release_date, months_of_support
        )
        self.assertEqual(end_year, 2012)
        self.assertEqual(end_month, 4)

        # test with modulo zero
        release_date = datetime.datetime(2010, 6, 10)
        (end_year, end_month) = get_maintenance_end_date(
            release_date, months_of_support
        )
        self.assertEqual(end_year, 2011)
        self.assertEqual(end_month, 12)

        # test dapper
        months_of_support = 60
        release_date = datetime.datetime(2008, 4, 24)
        (end_year, end_month) = get_maintenance_end_date(
            release_date, months_of_support
        )
        self.assertEqual(end_year, 2013)
        self.assertEqual(end_month, 4)

        # what datetime says
        # d = datetime.timedelta(18*30)
        # print "end date: ", release_date + d


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Michael Vogt <mvo@ubuntu.com>
#
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.
"""Unit tests for verifying the correctness of DebPackage in apt.debfile."""
import os
import logging
import unittest

from test_all import get_library_dir
import sys

libdir = get_library_dir()
if libdir:
    sys.path.insert(0, libdir)
import apt_pkg
import apt.debfile

import testcommon


class TestDebfile(testcommon.TestCase):
    """test the debfile"""

    TEST_DEBS = [
        # conflicts with apt
        ("gdebi-test1.deb", False),
        # impossible dependency
        ("gdebi-test2.deb", False),
        #  or-group (impossible-dependency|apt)
        ("gdebi-test3.deb", True),
        # Conflicts: apt (<= 0.1)
        ("gdebi-test4.deb", True),
        # Conflicts: apt (>= 0.1)
        ("gdebi-test5.deb", False),
        # invalid unicode  in descr
        ("gdebi-test6.deb", True),
        # provides/conflicts against "foobarbaz"
        ("gdebi-test7.deb", True),
        # provides/conflicts/replaces against "mail-transport-agent"
        # (should fails if mail-transport-agent is installed)
        ("gdebi-test8.deb", False),
        # provides/conflicts against real pkg
        ("gdebi-test9.deb", True),
        # provides debconf-tiny and the real debconf conflicts with
        ("gdebi-test10.deb", False),
    ]

    def setUp(self):
        testcommon.TestCase.setUp(self)
        apt_pkg.config.set("APT::Architecture", "i386")
        # FIXME: When run via test_all.py, the tests fail without this if it
        # is set in the system.
        apt_pkg.config.clear("APT::Architectures")
        apt_pkg.config.set("Dir::State::status", "./data/test_debs/var/lib/dpkg/status")
        apt_pkg.config.set("Dir::State::lists", "./data/test_debs/var/lib/apt/lists")
        apt_pkg.config.set(
            "Dir::Etc::sourcelist", "./data/test_debs/etc/apt/sources.list"
        )
        apt_pkg.init_system()
        self.cache = apt.Cache()

    def test_dsc_file(self):
        filename = "hello_2.5-1.dsc"
        deb = apt.debfile.DscSrcPackage(cache=self.cache)
        deb.open(os.path.join("data", "test_debs", filename))
        self.assertTrue(deb.check(), "got failure '%s'" % deb._failure_string)
        missing = set(["autotools-dev"])
        self.assertEqual(set(deb.missing_deps), missing)
        # specialized properties
        self.assertEqual(deb.pkgname, "hello")
        self.assertEqual(deb.binaries, ["hello", "bello", "cello"])
        self.assertEqual(deb.filelist, ["hello_2.5.orig.tar.gz", "hello_2.5-1.diff.gz"])
        self.assertEqual(deb.depends, [[("autotools-dev", "", "")]])
        # tag fields are available as a dict
        self.assertEqual(deb["Format"], "1.0")
        self.assertEqual(deb["Source"], "hello")
        self.assertEqual(deb["Binary"], "hello, bello,\n cello")
        self.assertEqual(deb["Architecture"], "any")
        self.assertEqual(deb["Version"], "2.5-1")
        self.assertEqual(deb["Maintainer"], "Santiago Vila <sanvila@debian.org>")
        self.assertEqual(deb["Homepage"], "http://www.gnu.org/software/hello")
        self.assertEqual(deb["Standards-Version"], "3.8.4")

    def test_dsc_file_with_impossible_build_dep(self):
        filename = "impossible-build-depends_2.5-1.dsc"
        deb = apt.debfile.DscSrcPackage(cache=self.cache)
        deb.open(os.path.join("data", "test_debs", filename))
        self.assertFalse(deb.check())
        self.assertEqual(deb.depends, [[("debhelper", "101", ">")]])

    def test_deb_file(self):
        deb = apt.debfile.DebPackage(cache=self.cache)
        for filename, expected_res in self.TEST_DEBS:
            logging.debug("testing %s, expecting %s" % (filename, expected_res))
            deb.open(os.path.join("data", "test_debs", filename))
            res = deb.check()
            self.assertEqual(
                res,
                expected_res,
                "Unexpected result for package '%s' (got %s wanted %s)\n%s"
                % (filename, res, expected_res, deb._failure_string),
            )

    def test_utf8_sections(self):
        deb = apt.debfile.DebPackage(cache=self.cache)
        deb.open(os.path.join("data", "test_debs", "utf8-package_1.0-1_all.deb"))
        self.assertEqual(deb["Maintainer"], "Samuel Lidén Borell <samuel@slbdata.se>")

    def test_content(self):
        # normal
        deb = apt.debfile.DebPackage(cache=self.cache)
        deb.open(os.path.join("data", "test_debs", "gdebi-test11.deb"))
        self.assertEqual('#!/bin/sh\necho "test"\n', deb.data_content("usr/bin/test"))
        # binary
        deb = apt.debfile.DebPackage(cache=self.cache)
        deb.open(os.path.join("data", "test_debs", "gdebi-test12.deb"))
        content = deb.data_content("usr/bin/binary")
        self.assertTrue(
            content.startswith("Automatically converted to printable ascii:\n\x7fELF ")
        )
        # control file
        needle = """Package: gdebi-test12
Version: 1.0
Architecture: all
Description: testpackage for gdebi - contains usr/bin/binary for file reading
 This tests the binary file reading for debfile.py
"""
        content = deb.control_content("./control")
        self.assertEqual(content, needle)
        content = deb.control_content("control")
        self.assertEqual(content, needle)

    def test_xz_data(self):
        deb = apt.debfile.DebPackage("./data/test_debs/data-tar-xz.deb")
        self.assertEqual(deb.filelist, ["./", "usr/", "usr/bin/"])

    @unittest.skipIf(
        apt_pkg.version_compare(apt_pkg.VERSION, "0.9.15.4~") < 0,
        "APT too old for uncompressed control.tar/data.tar",
    )
    def test_uncompressed_data(self):
        deb = apt.debfile.DebPackage("./data/test_debs/data-tar.deb")
        self.assertEqual(deb.filelist, ["./", "usr/", "usr/bin/"])

    def test_check_exception(self):
        deb = apt.debfile.DebPackage("./data/test_debs/data-tar-xz.deb")
        self.assertRaises(AttributeError, lambda: deb.missing_deps)
        deb.check()
        deb.missing_deps

    def test_no_supported_data_tar(self):
        with self.assertRaises(SystemError):
            apt.debfile.DebPackage("./data/test_debs/data-tar-broken.deb")

    def test_contains(self):
        deb = apt.debfile.DebPackage("./data/test_debs/data-tar-xz.deb")
        self.assertTrue("Package" in deb)

    def test_multi_arch_allowed(self):
        apt_pkg.config["APT::Architectures::"] = "i386"
        apt_pkg.config["APT::Architectures::"] = "amd64"
        apt_pkg.config["APT::Architecture"] = "amd64"
        apt_pkg.init_system()

        allowed_any = apt.debfile.DebPackage(
            "./data/test_debs/testdep-allowed-any_1.0-1_i386.deb"
        )
        self.assertTrue(allowed_any.check(), allowed_any._failure_string)

    def test_multi_arch_same(self):
        apt_pkg.config["APT::Architectures::"] = "i386"
        apt_pkg.config["APT::Architectures::"] = "amd64"
        apt_pkg.config["APT::Architecture"] = "amd64"
        apt_pkg.init_system()
        same = apt.debfile.DebPackage(
            "./data/test_debs/testdep-same-arch_1.0-1_i386.deb"
        )
        self.assertTrue(same.check(), same._failure_string)

    def test_get_content_gzip_data(self):
        deb = apt.debfile.DebPackage("./data/test_debs/gdebi-test13.deb")
        data = deb.data_content("./lala.gz")
        self.assertEqual(data, "Automatically decompressed:\n\nlala\n")


if __name__ == "__main__":
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Colomban Wendling <cwendling@hypra.fr>
#
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.
"""Unit tests for verifying InstallProgress works when spawning dpkg directly
(i.e. when installing standalone .debs)."""
import os
import unittest

from test_all import get_library_dir
import sys

libdir = get_library_dir()
if libdir:
    sys.path.insert(0, libdir)
from apt.progress.base import InstallProgress

import testcommon


class RunHelper:
    def __init__(self):
        self.script = "helper_install_progress_run.py"

    def do_install(self, fd):
        return os.spawnl(os.P_WAIT, self.script, self.script, str(fd))


class TestInstallProgressExec(testcommon.TestCase):
    """test that InstallProgress.run() passes a valid file descriptor to
    a child process"""

    def test_run(self):
        with InstallProgress() as prog:
            self.assertEqual(prog.run(RunHelper()), 0)


if __name__ == "__main__":
    os.chdir(os.path.dirname(__file__))
    unittest.main()



#!/usr/bin/env python2.4
#
# Test for the PkgSrcRecords code
# it segfaults for python-apt < 0.5.37
#

import apt_pkg
import sys


def main():
    apt_pkg.init()
    cache = apt_pkg.Cache()
    i = 0
    print("Running PkgSrcRecords test on all packages:")
    for x in cache.packages:
        i += 1
        src = apt_pkg.SourceRecords()
        if src.lookup(x.name):
            # print src.package
            pass
        print(
            "\r%i/%i=%.3f%%    "
            % (i, cache.package_count, (float(i) / float(cache.package_count) * 100))
        )


if __name__ == "__main__":
    main()
    sys.exit(0)



#!/usr/bin/python3
"""Emulate dpkg --contents"""

from __future__ import print_function

import grp
import pwd
import stat
import sys
import time

import apt_inst


def format_mode(member):
    """Return the symbolic mode"""
    mode = member.mode
    if member.isdir():
        s_mode = "d"
    elif member.islnk():
        s_mode = "h"
    else:
        s_mode = "-"
    s_mode += (mode & stat.S_IRUSR) and "r" or "-"
    s_mode += (mode & stat.S_IWUSR) and "w" or "-"
    s_mode += (
        (mode & stat.S_IXUSR)
        and (mode & stat.S_ISUID and "s" or "x")
        or (mode & stat.S_ISUID and "S" or "-")
    )
    s_mode += (mode & stat.S_IRGRP) and "r" or "-"
    s_mode += (mode & stat.S_IWGRP) and "w" or "-"
    s_mode += (
        (mode & stat.S_IXGRP)
        and (mode & stat.S_ISGID and "s" or "x")
        or (mode & stat.S_ISGID and "S" or "-")
    )
    s_mode += (mode & stat.S_IROTH) and "r" or "-"
    s_mode += (mode & stat.S_IWOTH) and "w" or "-"
    s_mode += (mode & stat.S_IXOTH) and "x" or "-"
    return s_mode


def callback(member, data):
    """callback for deb_extract"""
    s_mode = format_mode(member)
    s_owner = "%s/%s" % (pwd.getpwuid(member.uid)[0], grp.getgrgid(member.gid)[0])
    s_size = "%9d" % member.size
    s_time = time.strftime("%Y-%m-%d %H:%M", time.localtime(member.mtime))
    s_name = member.name if member.name.startswith(".") else ("./" + member.name)
    if member.islnk():
        s_name += " link to %s" % member.linkname
    print(s_mode, s_owner, s_size, s_time, s_name)


def main():
    """Main function"""
    if len(sys.argv) < 2:
        print("need filename argumnet", file=sys.stderr)
        sys.exit(1)

    fobj = open(sys.argv[1])
    try:
        apt_inst.DebFile(fobj).data.go(callback)
    finally:
        fobj.close()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
#
# Copyright (C) 2010 Julian Andres Klode <jak@debian.org>
#               2010 Michael Vogt <mvo@ubuntu.com>
#
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.
"""Unit tests for verifying the correctness of check_dep, etc in apt_pkg."""

from __future__ import print_function

import glob
import logging
import os
import shutil
import sys
import tempfile
import unittest


from test_all import get_library_dir

libdir = get_library_dir()
if libdir:
    sys.path.insert(0, libdir)

import apt
import apt_pkg

import testcommon


def if_sources_list_is_readable(f):
    def wrapper(*args, **kwargs):
        if os.access("/etc/apt/sources.list", os.R_OK):
            f(*args, **kwargs)
        else:
            logging.warning("skipping '%s' because sources.list is not readable" % f)

    return wrapper


def get_open_file_descriptors():
    try:
        fds = os.listdir("/proc/self/fd")
    except OSError:
        logging.warning("failed to list /proc/self/fd")
        return set([])
    return set(map(int, fds))


class TestAptCache(testcommon.TestCase):
    """test the apt cache"""

    def setUp(self):
        testcommon.TestCase.setUp(self)
        apt_pkg.config.clear("APT::Update::Post-Invoke")
        apt_pkg.config.clear("APT::Update::Post-Invoke-Success")

    @if_sources_list_is_readable
    def test_apt_cache(self):
        """cache: iterate all packages and all dependencies"""
        cache = apt.Cache()
        # number is not meaningful and just need to be "big enough",
        # the important bit is the test against __len__
        self.assertTrue(len(cache) > 100)
        # go over the cache and all dependencies, just to see if
        # that is possible and does not crash
        for pkg in cache:
            if pkg.candidate:
                for or_deps in pkg.candidate.dependencies:
                    for dep in or_deps:
                        self.assertTrue(dep.name)
                        self.assertTrue(isinstance(dep.relation, str))
                        self.assertTrue(dep.pre_depend in (True, False))

                # accessing record should take a reasonable time; in
                # particular, when using compressed indexes, it should not use
                # tons of seek operations
                r = pkg.candidate.record
                self.assertEqual(r["Package"], pkg.shortname)
                self.assertTrue("Version" in r)
                self.assertTrue(len(r["Description"]) > 0)
                self.assertTrue(str(r).startswith("Package: %s\n" % pkg.shortname))

    @if_sources_list_is_readable
    def test_cache_close_leak_fd(self):
        fds_before_open = get_open_file_descriptors()
        cache = apt.Cache()
        opened_fd = get_open_file_descriptors().difference(fds_before_open)
        cache.close()
        fds_after_close = get_open_file_descriptors()
        unclosed_fd = opened_fd.intersection(fds_after_close)
        self.assertEqual(fds_before_open, fds_after_close)
        self.assertEqual(unclosed_fd, set())

    def test_cache_open_twice_leaks_fds(self):
        cache = apt.Cache()
        fds_before_open = get_open_file_descriptors()
        cache.open()
        fds_after_open_twice = get_open_file_descriptors()
        self.assertEqual(fds_before_open, fds_after_open_twice)

    @if_sources_list_is_readable
    def test_cache_close_download_fails(self):
        cache = apt.Cache()
        self.assertEqual(cache.required_download, 0)
        cache.close()
        with self.assertRaises(apt.cache.CacheClosedException):
            cache.required_download

    def test_get_provided_packages(self):
        apt.apt_pkg.config.set("Apt::architecture", "i386")
        cache = apt.Cache(rootdir="./data/test-provides/")
        cache.open()
        if len(cache) == 0:
            logging.warning("skipping test_get_provided_packages, cache empty?!?")
            return
        # a true virtual pkg
        li = cache.get_providing_packages("mail-transport-agent")
        self.assertTrue(len(li) > 0)
        self.assertTrue("postfix" in [p.name for p in li])
        self.assertTrue("mail-transport-agent" in cache["postfix"].candidate.provides)

    def test_low_level_pkg_provides(self):
        apt.apt_pkg.config.set("Apt::architecture", "i386")
        # create highlevel cache and get the lowlevel one from it
        highlevel_cache = apt.Cache(rootdir="./data/test-provides")
        if len(highlevel_cache) == 0:
            logging.warning("skipping test_log_level_pkg_provides, cache empty?!?")
            return
        # low level cache provides list of the pkg
        cache = highlevel_cache._cache
        li = cache["mail-transport-agent"].provides_list
        # arbitrary number, just needs to be higher enough
        self.assertEqual(len(li), 2)
        for providesname, providesver, version in li:
            self.assertEqual(providesname, "mail-transport-agent")
            if version.parent_pkg.name == "postfix":
                break
        else:
            self.assertNotReached()

    @if_sources_list_is_readable
    def test_dpkg_journal_dirty(self):
        # create tmp env
        tmpdir = tempfile.mkdtemp()
        dpkg_dir = os.path.join(tmpdir, "var", "lib", "dpkg")
        os.makedirs(os.path.join(dpkg_dir, "updates"))
        open(os.path.join(dpkg_dir, "status"), "w").close()
        apt_pkg.config.set("Dir::State::status", os.path.join(dpkg_dir, "status"))
        cache = apt.Cache()
        # test empty
        self.assertFalse(cache.dpkg_journal_dirty)
        # that is ok, only [0-9] are dpkg jounral entries
        open(os.path.join(dpkg_dir, "updates", "xxx"), "w").close()
        self.assertFalse(cache.dpkg_journal_dirty)
        # that is a dirty journal
        open(os.path.join(dpkg_dir, "updates", "000"), "w").close()
        self.assertTrue(cache.dpkg_journal_dirty)

    @if_sources_list_is_readable
    def test_apt_update(self):
        rootdir = "./data/tmp"
        if os.path.exists(rootdir):
            shutil.rmtree(rootdir)
        try:
            os.makedirs(os.path.join(rootdir, "var/lib/apt/lists/partial"))
        except OSError:
            pass
        state_dir = os.path.join(rootdir, "var/lib/apt")
        lists_dir = os.path.join(rootdir, "var/lib/apt/lists")
        old_state = apt_pkg.config.find("dir::state")
        apt_pkg.config.set("dir::state", state_dir)
        # set a local sources.list that does not need the network
        base_sources = os.path.abspath(os.path.join(rootdir, "sources.list"))
        old_source_list = apt_pkg.config.find("dir::etc::sourcelist")
        old_source_parts = apt_pkg.config.find("dir::etc::sourceparts")
        apt_pkg.config.set("dir::etc::sourcelist", base_sources)
        # TODO: /dev/null is not a dir, perhaps find something better
        apt_pkg.config.set("dir::etc::sourceparts", "/dev/null")
        # main sources.list
        sources_list = base_sources
        with open(sources_list, "w") as f:
            repo = os.path.abspath("./data/test-repo2")
            f.write("deb [allow-insecure=yes] copy:%s /\n" % repo)

        # test single sources.list fetching
        sources_list = os.path.join(rootdir, "test.list")
        with open(sources_list, "w") as f:
            repo_dir = os.path.abspath("./data/test-repo")
            f.write("deb [allow-insecure=yes] copy:%s /\n" % repo_dir)

        self.assertTrue(os.path.exists(sources_list))
        # write marker to ensure listcleaner is not run
        open("./data/tmp/var/lib/apt/lists/marker", "w").close()

        # update a single sources.list
        cache = apt.Cache()
        cache.update(sources_list=sources_list)
        # verify we just got the excpected package file
        needle_packages = glob.glob(lists_dir + "/*tests_data_test-repo_Packages*")
        self.assertEqual(len(needle_packages), 1)
        # verify that we *only* got the Packages file from a single source
        all_packages = glob.glob(lists_dir + "/*_Packages*")
        self.assertEqual(needle_packages, all_packages)
        # verify that the listcleaner was not run and the marker file is
        # still there
        self.assertTrue("marker" in os.listdir(lists_dir))
        # now run update again (without the "normal" sources.list that
        # contains test-repo2 and verify that we got the normal sources.list
        cache.update()
        needle_packages = glob.glob(lists_dir + "/*tests_data_test-repo2_Packages*")
        self.assertEqual(len(needle_packages), 1)
        all_packages = glob.glob(lists_dir + "/*_Packages*")
        self.assertEqual(needle_packages, all_packages)

        # and another update with a single source only
        cache = apt.Cache()
        cache.update(sources_list=sources_list)
        all_packages = glob.glob(lists_dir + "/*_Packages*")
        self.assertEqual(len(all_packages), 2)
        apt_pkg.config.set("dir::state", old_state)
        apt_pkg.config.set("dir::etc::sourcelist", old_source_list)
        apt_pkg.config.set("dir::etc::sourceparts", old_source_parts)

    def test_package_cmp(self):
        cache = apt.Cache(rootdir="/")
        li = []
        li.append(cache["intltool"])
        li.append(cache["python3"])
        li.append(cache["apt"])
        li.sort()
        self.assertEqual([p.name for p in li], ["apt", "intltool", "python3"])

    def test_get_architectures(self):
        main_arch = apt.apt_pkg.config.get("APT::Architecture")
        arches = apt_pkg.get_architectures()
        self.assertTrue(main_arch in arches)

    def test_apt_cache_reopen_is_safe(self):
        """cache: check that we cannot use old package objects after reopen"""
        cache = apt.Cache()
        old_depcache = cache._depcache
        old_package = cache["apt"]
        old_pkg = old_package._pkg
        old_version = old_package.candidate
        old_ver = old_version._cand

        cache.open()
        new_depcache = cache._depcache
        new_package = cache["apt"]
        new_pkg = new_package._pkg
        new_version = new_package.candidate
        new_ver = new_version._cand

        # get candidate
        self.assertRaises(ValueError, old_depcache.get_candidate_ver, new_pkg)
        self.assertRaises(ValueError, new_depcache.get_candidate_ver, old_pkg)
        self.assertEqual(new_ver, new_depcache.get_candidate_ver(new_pkg))
        self.assertEqual(old_package.candidate._cand, old_ver)  # Remap success
        self.assertEqual(
            old_package.candidate._cand, new_depcache.get_candidate_ver(new_pkg)
        )

        # set candidate
        new_package.candidate = old_version
        old_depcache.set_candidate_ver(old_pkg, old_ver)
        self.assertRaises(ValueError, old_depcache.set_candidate_ver, old_pkg, new_ver)
        self.assertRaises(ValueError, new_depcache.set_candidate_ver, old_pkg, old_ver)
        self.assertRaises(ValueError, new_depcache.set_candidate_ver, old_pkg, new_ver)
        self.assertRaises(ValueError, new_depcache.set_candidate_ver, new_pkg, old_ver)
        new_depcache.set_candidate_ver(new_pkg, new_ver)

    @staticmethod
    def write_status_file(packages):
        with open(apt_pkg.config["Dir::State::Status"], "w") as fobj:
            for package in packages:
                print("Package:", package, file=fobj)
                print("Status: install ok installed", file=fobj)
                print("Priority: optional", file=fobj)
                print("Section: admin", file=fobj)
                print("Installed-Size: 1", file=fobj)
                print("Maintainer: X <x@x.invalid>", file=fobj)
                print("Architecture: all", file=fobj)
                print("Version: 1", file=fobj)
                print("Description: blah", file=fobj)
                print("", file=fobj)

    def test_apt_cache_reopen_is_safe_out_of_bounds(self):
        """Check that out of bounds access is remapped correctly."""
        with tempfile.NamedTemporaryFile() as status:
            apt_pkg.config["Dir::Etc::SourceList"] = "/dev/null"
            apt_pkg.config["Dir::Etc::SourceParts"] = "/dev/null"
            apt_pkg.config["Dir::State::Status"] = status.name
            apt_pkg.init_system()

            self.write_status_file("abcdefghijklmnopqrstuvwxyz")
            c = apt.Cache()
            p = c["z"]
            p_id = p.id
            self.write_status_file("az")
            apt_pkg.init_system()
            c.open()
            self.assertNotEqual(p.id, p_id)
            self.assertLess(p.id, 2)
            p.mark_delete()
            self.assertEqual([p], c.get_changes())

    def test_apt_cache_reopen_is_safe_out_of_bounds_no_match(self):
        """Check that installing gone package raises correct exception"""
        with tempfile.NamedTemporaryFile() as status:
            apt_pkg.config["Dir::Etc::SourceList"] = "/dev/null"
            apt_pkg.config["Dir::Etc::SourceParts"] = "/dev/null"
            apt_pkg.config["Dir::State::Status"] = status.name
            apt_pkg.init_system()

            self.write_status_file("abcdefghijklmnopqrstuvwxyz")
            c = apt.Cache()
            p = c["z"]
            p_id = p.id
            self.write_status_file("a")
            apt_pkg.init_system()
            c.open()
            self.assertEqual(p.id, p_id)  # Could not be remapped
            self.assertRaises(apt_pkg.CacheMismatchError, p.mark_delete)

    def test_apt_cache_reopen_is_safe_swap(self):
        """Check that swapping a and b does not mark the wrong package."""
        with tempfile.NamedTemporaryFile() as status:
            apt_pkg.config["Dir::Etc::SourceList"] = "/dev/null"
            apt_pkg.config["Dir::Etc::SourceParts"] = "/dev/null"
            apt_pkg.config["Dir::State::Status"] = status.name
            apt_pkg.init_system()

            self.write_status_file("abcdefghijklmnopqrstuvwxyz")
            c = apt.Cache()
            p = c["a"]
            a_id = p.id
            p_hash = hash(p)
            set_of_p = set([p])
            self.write_status_file("baz")
            apt_pkg.init_system()
            c.open()
            # b now has the same id as a in the old cache
            self.assertEqual(c["b"].id, a_id)
            self.assertNotEqual(p.id, a_id)
            # Marking a should still mark a, not b.
            p.mark_delete()
            self.assertEqual([p], c.get_changes())

            # Ensure that p can still be found in a set of it, as a test
            # for bug https://bugs.launchpad.net/bugs/1780099
            self.assertEqual(hash(p), p_hash)
            self.assertIn(p, set_of_p)

    def test_apt_cache_iteration_safe(self):
        """Check that iterating does not produce different results.

        This failed in 1.7.0~alpha2, because one part of the code
        looked up packages in the weak dict using the pretty name,
        and the other using the full name."""

        with tempfile.NamedTemporaryFile() as status:
            apt_pkg.config["Dir::Etc::SourceList"] = "/dev/null"
            apt_pkg.config["Dir::Etc::SourceParts"] = "/dev/null"
            apt_pkg.config["Dir::State::Status"] = status.name
            apt_pkg.init_system()

            self.write_status_file("abcdefghijklmnopqrstuvwxyz")

            c = apt.Cache()
            c["a"].mark_delete()
            self.assertEqual([c["a"]], [p for p in c if p.marked_delete])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python2.4
#
# Test for the DepCache code
#

import apt_pkg
import sys


def main():
    apt_pkg.init()
    cache = apt_pkg.Cache()
    depcache = apt_pkg.DepCache(cache)
    depcache.init()
    i = 0
    all = cache.package_count
    print("Running DepCache test on all packages")
    print("(trying to install each and then mark it keep again):")
    # first, get all pkgs
    for pkg in cache.packages:
        i += 1
        x = pkg.name
        # then get each version
        ver = depcache.get_candidate_ver(pkg)
        if ver is not None:
            depcache.mark_install(pkg)
            if depcache.inst_count == 0:
                if depcache.is_upgradable(pkg):
                    print("Error marking %s for install" % x)
            for p in cache.packages:
                if depcache.marked_install(p):
                    depcache.mark_keep(p)
            if depcache.inst_count != 0:
                print(
                    "Error undoing the selection for %s (inst_count: %s)"
                    % (x, depcache.inst_count)
                )
        print("\r%i/%i=%.3f%%    " % (i, all, (float(i) / float(all) * 100)))

    print()
    print("Trying upgrade:")
    depcache.upgrade()
    print("To install: %s " % depcache.inst_count)
    print("To remove: %s " % depcache.del_count)
    print("Kept back: %s " % depcache.keep_count)

    print("Trying DistUpgrade:")
    depcache.upgrade(True)
    print("To install: %s " % depcache.inst_count)
    print("To remove: %s " % depcache.del_count)
    print("Kept back: %s " % depcache.keep_count)


if __name__ == "__main__":
    main()
    sys.exit(0)



#!/usr/bin/python3

from __future__ import print_function

import contextlib
import errno
import itertools
import os
import shutil
import sys
import tempfile
import time
import unittest

from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler as HTTPRequestHandler

import apt_pkg
import apt.auth

import testcommon

WHEEZY_KEYID = "8B48AD6246925553"
WHEEZY_KEYDATE = "1335553717"
WHEEZY_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1.4.12 (GNU/Linux)


DHcut3Yey8o=
=id4q
-----END PGP PUBLIC KEY BLOCK-----"""


def normalize_key(keystr):
    """Remove the Version: header from a key block"""
    lines = keystr.split("\n")
    if lines[1].startswith("Version:"):
        return lines[:1] + lines[2:]
    return lines


class TestAuthKeys(testcommon.TestCase):

    """Test handling of keys for signed repositories."""

    def setUp(self):
        testcommon.TestCase.setUp(self)

        self.tmpdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmpdir)
        apt_pkg.config.set("Dir", self.tmpdir)
        apt_pkg.config.set("Dir::Bin::Apt-key", "fakeroot-apt-key")
        apt_pkg.config.set("Dir::Etc", "etc/apt/")
        trustedparts_dir = apt_pkg.config.find_dir("Dir::Etc::Trustedparts")
        confparts_dir = apt_pkg.config.find_dir("Dir::Etc::parts")
        self.assertTrue(trustedparts_dir.startswith(self.tmpdir))
        os.makedirs(trustedparts_dir)
        os.makedirs(confparts_dir)
        shutil.copy("fakeroot-apt-key", self.tmpdir)

    @contextlib.contextmanager
    def _discard_stderr(self):
        stderr_fd = sys.stderr.fileno()
        stderr_save = os.dup(stderr_fd)
        try:
            devnull = os.open("/dev/null", os.O_WRONLY)
            try:
                os.dup2(devnull, stderr_fd)
                yield
            finally:
                os.close(devnull)
        finally:
            os.dup2(stderr_save, stderr_fd)
            os.close(stderr_save)

    def testAddAndExportKey(self):
        """Add an example key."""
        apt.auth.add_key(WHEEZY_KEY)
        # Strip the headers from the keys to avoid test errors because
        # the exported key used a differenct GnuPG version than the
        # original example key
        self.assertEqual(
            normalize_key(apt.auth.export_key(WHEEZY_KEYID)), normalize_key(WHEEZY_KEY)
        )

    def testAddAndListKey(self):
        """Add an example key and test if it is correctly returned by
        list_keys()
        """
        apt.auth.add_key(WHEEZY_KEY)
        ret = apt.auth.list_keys()
        self.assertEqual(len(ret), 1)
        key = ret[0]
        self.assertEqual(
            key.name,
            "Debian Archive Automatic Signing Key (7.0/wheezy) "
            "<ftpmaster@debian.org>",
        )
        self.assertEqual(key.keyid, WHEEZY_KEYID)
        self.assertEqual(key.date, WHEEZY_KEYDATE)

    def testAddKeyFromFile(self):
        """Test adding a key from file."""
        keyfd, keyname = tempfile.mkstemp()
        self.addCleanup(os.close, keyfd)
        os.write(keyfd, WHEEZY_KEY.encode("UTF-8"))

        apt.auth.add_key_from_file(keyname)

        ret = apt.auth.list_keys()
        self.assertEqual(len(ret), 1)
        key = ret[0]
        self.assertEqual(
            key.name,
            "Debian Archive Automatic Signing Key (7.0/wheezy) "
            "<ftpmaster@debian.org>",
        )
        self.assertEqual(key.keyid, WHEEZY_KEYID)
        self.assertEqual(key.date, WHEEZY_KEYDATE)

    def test_add_key_from_keyserver_too_short(self):
        """Ensure that short keyids are not imported"""
        with self.assertRaises(apt.auth.AptKeyIDTooShortError):
            apt.auth.add_key_from_keyserver(WHEEZY_KEYID, "hkp://localhost:19191")
        with self.assertRaises(apt.auth.AptKeyIDTooShortError):
            apt.auth.add_key_from_keyserver(
                "0101010178F7FE5C3E65D8AF8B48AD624692555", "hkp://localhost:19191"
            )
        with self.assertRaises(apt.auth.AptKeyIDTooShortError):
            apt.auth.add_key_from_keyserver(
                "0x0101010178F7FE5C3E65D8AF8B48AD624692555", "hkp://localhost:19191"
            )
        with self.assertRaises(apt.auth.AptKeyIDTooShortError):
            apt.auth.add_key_from_keyserver(
                "0101 0101 78F7 FE5C 3E65 D8AF 8B48 AD62 4692 555",
                "hkp://localhost:19191",
            )

    def test_add_key_from_server_mitm(self):
        """Verify that the key fingerprint is verified after download"""
        self._start_keyserver()
        self.addCleanup(self._stop_keyserver)
        with self.assertRaises(apt.auth.AptKeyError) as cm:
            with self._discard_stderr():
                apt.auth.add_key_from_keyserver(
                    "0101010178F7FE5C3E65D8AF8B48AD6246925553",
                    "hkp://localhost:%d" % self.keyserver_port,
                )
        self.assertTrue(
            str(cm.exception).startswith(
                "recv from 'hkp://localhost:%d' failed for '%s'"
                % (self.keyserver_port, "0101010178F7FE5C3E65D8AF8B48AD6246925553")
            ),
            cm.exception,
        )

    def testAddKeyFromServer(self):
        """Install a GnuPG key from a remote server."""
        self._start_keyserver()
        self.addCleanup(self._stop_keyserver)

        with self._discard_stderr():
            apt.auth.add_key_from_keyserver(
                "0xa1bD8E9D78F7FE5C3E65D8AF8B48AD6246925553",
                "hkp://localhost:%d" % self.keyserver_port,
            )

        ret = apt.auth.list_keys()
        self.assertEqual(len(ret), 1)
        key = ret[0]
        self.assertEqual(
            key.name,
            "Debian Archive Automatic Signing Key (7.0/wheezy) "
            "<ftpmaster@debian.org>",
        )
        self.assertEqual(key.keyid, WHEEZY_KEYID)
        self.assertEqual(key.date, WHEEZY_KEYDATE)

    def _start_keyserver(self):
        """Start a fake keyserver on http://localhost:19191
        If port 19191 is unavailable, try successive ports until one is.
        Store the port actually in use in self.keyserver_port.
        Thanks pitti.
        """
        dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, dir)
        os.mkdir(os.path.join(dir, "pks"))
        with open(os.path.join(dir, "pks", "lookup"), "w") as key_file:
            key_file.write(WHEEZY_KEY)

        keyserver_pipe = os.pipe()
        self.keyserver_pid = os.fork()
        if self.keyserver_pid == 0:
            os.close(keyserver_pipe[0])
            # quiesce server log
            os.dup2(os.open("/dev/null", os.O_WRONLY), sys.stderr.fileno())
            os.chdir(dir)
            for port in itertools.count(19191):
                try:
                    httpd = HTTPServer(("localhost", port), HTTPRequestHandler)
                    break
                except IOError as e:
                    if e.errno != errno.EADDRINUSE:
                        raise
            keyserver_write = os.fdopen(keyserver_pipe[1], "w")
            print(port, file=keyserver_write)
            keyserver_write.close()
            httpd.serve_forever()
            os._exit(0)

        os.close(keyserver_pipe[1])
        keyserver_read = os.fdopen(keyserver_pipe[0])
        self.keyserver_port = int(keyserver_read.readline())
        keyserver_read.close()

        # temporarily disable proxy, as gnupg does not get along with that
        # (LP #789049)
        self.orig_proxy = os.environ.get("http_proxy")
        try:
            del os.environ["http_proxy"]
        except KeyError:
            pass

        # wait a bit until server is ready
        time.sleep(0.5)

    def _stop_keyserver(self):
        """Stop fake keyserver"""
        assert self.keyserver_pid

        os.kill(self.keyserver_pid, 15)
        os.wait()

        # restore proxy
        if self.orig_proxy is not None:
            os.environ["http_proxy"] = self.orig_proxy


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# SPDX-License-Identifier: GPL-2.0+


import base64
import os
import shutil
import tempfile
import unittest

import apt_pkg
import apt
import apt.progress.base
import apt.progress.text

import testcommon

# A change in APT to early fail acquire items with weak hashes caused it
# to call progress methods before calling Start(), which confused python-apt's
# locking handling. Hence we need to skip such tests for now, until we can test
# against an apt version with the fix:
#
#     https://salsa.debian.org/apt-team/apt/commit/84176f6c
#
RUN_CRASHING_TESTS = False

# Message APT gives us when hashes are too weak
CACHE_MSG_WEAK_HASH = (
    "Insufficient information available to perform this download securely"
)


class TestSignedUsable(testcommon.TestCase):
    """Test fetch_binary() and fetch_source() signature checking."""

    def setUp(self):
        testcommon.TestCase.setUp(self)
        apt_pkg.config.clear("APT::Update::Post-Invoke")
        apt_pkg.config.clear("APT::Update::Post-Invoke-Success")
        self.chroot_path = chroot_path = tempfile.mkdtemp()
        repo_path = os.path.abspath("./data/test-signed-usable-repo/")
        # Inits the dirs for us
        apt.cache.Cache(rootdir=chroot_path)
        # Change directory
        self.cwd = os.getcwd()
        os.chdir(chroot_path)
        with open(
            os.path.join(self.chroot_path, "etc/apt/sources.list"), "w"
        ) as sources_list:
            sources_list.write("deb copy:%s/signed/ /\n" % repo_path)
            sources_list.write("deb copy:%s/unsigned/ /\n" % repo_path)
            sources_list.write("deb-src copy:%s/signed/ /\n" % repo_path)
            sources_list.write("deb-src copy:%s/unsigned/ /\n" % repo_path)

        with open(os.path.join(repo_path, "key.gpg.base64"), "rb") as pubkey64:
            with open(
                os.path.join(self.chroot_path, "etc/apt/trusted.gpg"), "wb"
            ) as tgt:
                tgt.write(base64.b64decode(pubkey64.read()))

        self.cache = apt.cache.Cache(rootdir=chroot_path)
        apt_pkg.config["Acquire::AllowInsecureRepositories"] = "true"
        self.cache.update()
        apt_pkg.config["Acquire::AllowInsecureRepositories"] = "false"
        self.cache.open()

        self.progress = apt.progress.text.AcquireProgress
        apt.progress.text.AcquireProgress = apt.progress.base.AcquireProgress

        # Disable actual installation of downloaded items
        self.cache.install_archives = (
            lambda *a, **b: apt_pkg.PackageManager.RESULT_COMPLETED
        )

    def tearDown(self):
        # this resets the rootdir apt_pkg.config to ensure it does not
        # "pollute" the later tests
        apt.cache.Cache(rootdir="/")
        os.chdir(self.cwd)
        shutil.rmtree(self.chroot_path)

        apt.progress.text.AcquireProgress = self.progress

    def doInstall(self, name, bargs):
        self.cache[name].mark_install()
        try:
            with apt.progress.base.InstallProgress() as ip:
                self.cache.commit(install_progress=ip, **bargs)
        finally:
            for fname in os.listdir(
                os.path.join(self.chroot_path, "var/cache/apt/archives")
            ):
                if os.path.isfile(
                    os.path.join(self.chroot_path, "var/cache/apt/archives", fname)
                ):
                    os.unlink(
                        os.path.join(self.chroot_path, "var/cache/apt/archives", fname)
                    )
            self.cache[name].mark_keep()

    def doFetchArchives(self, name, bargs):
        fetcher = apt_pkg.Acquire()
        self.cache[name].mark_install()
        try:
            self.cache.fetch_archives(fetcher=fetcher, **bargs)
        finally:
            for fname in os.listdir(
                os.path.join(self.chroot_path, "var/cache/apt/archives")
            ):
                if fname.endswith(".deb"):
                    os.unlink(
                        os.path.join(self.chroot_path, "var/cache/apt/archives", fname)
                    )
            self.cache[name].mark_keep()

    def testDefaultDenyButExplicitAllowUnauthenticated(self):
        """Deny by config (default), but pass allow_unauthenticated=True"""

        bargs = dict(allow_unauthenticated=True)
        sargs = dict(allow_unauthenticated=True, unpack=False)

        self.doInstall("signed-usable", bargs)
        if RUN_CRASHING_TESTS:
            self.assertRaisesRegex(
                apt.cache.FetchFailedException,
                CACHE_MSG_WEAK_HASH,
                self.doInstall,
                "signed-not-usable",
                bargs,
            )

        self.doInstall("unsigned-usable", bargs)
        if RUN_CRASHING_TESTS:
            self.assertRaisesRegex(
                apt.cache.FetchFailedException,
                CACHE_MSG_WEAK_HASH,
                self.doInstall,
                "unsigned-unusable",
                bargs,
            )

        self.doFetchArchives("signed-usable", bargs)
        self.doFetchArchives("unsigned-usable", bargs)
        self.assertRaisesRegex(
            apt.cache.FetchFailedException,
            CACHE_MSG_WEAK_HASH,
            self.doFetchArchives,
            "signed-not-usable",
            bargs,
        )
        self.assertRaisesRegex(
            apt.cache.FetchFailedException,
            CACHE_MSG_WEAK_HASH,
            self.doFetchArchives,
            "unsigned-unusable",
            bargs,
        )

        self.cache["signed-usable"].candidate.fetch_binary(**bargs)
        self.cache["signed-usable"].candidate.fetch_source(**sargs)
        self.cache["signed-not-usable"].candidate.fetch_binary(**bargs)
        self.cache["signed-not-usable"].candidate.fetch_source(**sargs)
        self.cache["unsigned-usable"].candidate.fetch_binary(**bargs)
        self.cache["unsigned-usable"].candidate.fetch_source(**sargs)
        self.cache["unsigned-unusable"].candidate.fetch_binary(**bargs)
        self.cache["unsigned-unusable"].candidate.fetch_source(**sargs)

    def testDefaultAllow(self):
        """Allow by config APT::Get::AllowUnauthenticated = True"""
        apt_pkg.config["APT::Get::AllowUnauthenticated"] = "true"

        bargs = dict()
        sargs = dict(unpack=False)

        self.doInstall("signed-usable", bargs)
        if RUN_CRASHING_TESTS:
            self.assertRaisesRegex(
                apt.cache.FetchFailedException,
                CACHE_MSG_WEAK_HASH,
                self.doInstall,
                "signed-not-usable",
                bargs,
            )
        self.doInstall("unsigned-usable", bargs)
        if RUN_CRASHING_TESTS:
            self.assertRaisesRegex(
                apt.cache.FetchFailedException,
                CACHE_MSG_WEAK_HASH,
                self.doInstall,
                "unsigned-unusable",
                bargs,
            )

        self.doFetchArchives("signed-usable", bargs)
        self.doFetchArchives("unsigned-usable", bargs)
        self.assertRaisesRegex(
            apt.cache.FetchFailedException,
            CACHE_MSG_WEAK_HASH,
            self.doFetchArchives,
            "signed-not-usable",
            bargs,
        )
        self.assertRaisesRegex(
            apt.cache.FetchFailedException,
            CACHE_MSG_WEAK_HASH,
            self.doFetchArchives,
            "unsigned-unusable",
            bargs,
        )

        self.cache["signed-usable"].candidate.fetch_binary(**bargs)
        self.cache["signed-usable"].candidate.fetch_source(**sargs)
        self.cache["signed-not-usable"].candidate.fetch_binary(**bargs)
        self.cache["signed-not-usable"].candidate.fetch_source(**sargs)
        self.cache["unsigned-usable"].candidate.fetch_binary(**bargs)
        self.cache["unsigned-usable"].candidate.fetch_source(**sargs)
        self.cache["unsigned-unusable"].candidate.fetch_binary(**bargs)
        self.cache["unsigned-unusable"].candidate.fetch_source(**sargs)

    def testDefaultDeny(self):
        """Test APT::Get::AllowUnauthenticated = False (default)"""
        self.doInstall("signed-usable", {})
        if RUN_CRASHING_TESTS:
            self.assertRaisesRegex(
                apt.cache.FetchFailedException,
                CACHE_MSG_WEAK_HASH,
                self.doInstall,
                "signed-not-usable",
                {},
            )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doInstall,
            "unsigned-usable",
            {},
        )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doInstall,
            "unsigned-unusable",
            {},
        )

        self.doFetchArchives("signed-usable", {})
        self.assertRaisesRegex(
            apt.cache.FetchFailedException,
            CACHE_MSG_WEAK_HASH,
            self.doFetchArchives,
            "signed-not-usable",
            {},
        )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doFetchArchives,
            "unsigned-usable",
            {},
        )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doFetchArchives,
            "unsigned-unusable",
            {},
        )

        self.cache["signed-usable"].candidate.fetch_binary()
        self.cache["signed-usable"].candidate.fetch_source(unpack=False)
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": No trusted hash",
            self.cache["signed-not-usable"].candidate.fetch_binary,
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": No trusted hash",
            self.cache["signed-not-usable"].candidate.fetch_source,
            unpack=False,
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-usable"].candidate.fetch_binary,
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-usable"].candidate.fetch_source,
            unpack=False,
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-unusable"].candidate.fetch_binary,
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-unusable"].candidate.fetch_source,
            unpack=False,
        )

    def testDefaultAllowButExplicitDeny(self):
        """Allow by config, but pass allow_unauthenticated=False"""
        apt_pkg.config["APT::Get::AllowUnauthenticated"] = "true"

        bargs = dict(allow_unauthenticated=False)
        sargs = dict(allow_unauthenticated=False, unpack=False)

        self.doInstall("signed-usable", bargs)
        if RUN_CRASHING_TESTS:
            self.assertRaisesRegex(
                apt.cache.FetchFailedException,
                CACHE_MSG_WEAK_HASH,
                self.doInstall,
                "signed-not-usable",
                bargs,
            )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doInstall,
            "unsigned-usable",
            bargs,
        )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doInstall,
            "unsigned-unusable",
            bargs,
        )

        self.doFetchArchives("signed-usable", bargs)
        self.assertRaisesRegex(
            apt.cache.FetchFailedException,
            CACHE_MSG_WEAK_HASH,
            self.doFetchArchives,
            "signed-not-usable",
            bargs,
        )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doFetchArchives,
            "unsigned-usable",
            bargs,
        )
        self.assertRaisesRegex(
            apt.cache.UntrustedException,
            "Untrusted packages:",
            self.doFetchArchives,
            "unsigned-unusable",
            bargs,
        )

        self.cache["signed-usable"].candidate.fetch_binary(**bargs)
        self.cache["signed-usable"].candidate.fetch_source(**sargs)
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": No trusted hash",
            self.cache["signed-not-usable"].candidate.fetch_binary,
            **bargs
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": No trusted hash",
            self.cache["signed-not-usable"].candidate.fetch_source,
            **sargs
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-usable"].candidate.fetch_binary,
            **bargs
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-usable"].candidate.fetch_source,
            **sargs
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-unusable"].candidate.fetch_binary,
            **bargs
        )
        self.assertRaisesRegex(
            apt.package.UntrustedError,
            ": Source",
            self.cache["unsigned-unusable"].candidate.fetch_source,
            **sargs
        )


if __name__ == "__main__":
    unittest.main()



# package.py - apt package abstraction
#
#  Copyright (c) 2005-2009 Canonical
#
#  Author: Michael Vogt <michael.vogt@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
"""Functionality related to packages."""
from __future__ import print_function

import logging
import os
import sys
import re
import socket
import subprocess
import threading

from http.client import BadStatusLine
from urllib.error import HTTPError
from urllib.request import urlopen

from typing import (
    Any,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    no_type_check,
    Mapping,
    Sequence,
)

import apt_pkg
import apt.progress.text

from apt.progress.base import (
    AcquireProgress,
    InstallProgress,
)

from apt_pkg import gettext as _

__all__ = (
    "BaseDependency",
    "Dependency",
    "Origin",
    "Package",
    "Record",
    "Version",
    "VersionList",
)


def _file_is_same(path, size, hashes):
    # type: (str, int, apt_pkg.HashStringList) -> bool
    """Return ``True`` if the file is the same."""
    if os.path.exists(path) and os.path.getsize(path) == size:
        with open(path) as fobj:
            return apt_pkg.Hashes(fobj).hashes == hashes
    return False


class FetchError(Exception):
    """Raised when a file could not be fetched."""


class UntrustedError(FetchError):
    """Raised when a file did not have a trusted hash."""


class BaseDependency(object):
    """A single dependency."""

    class __dstr(str):
        """Compare helper for compatibility with old third-party code.

        Old third-party code might still compare the relation with the
        previously used relations (<<,<=,==,!=,>=,>>,) instead of the curently
        used ones (<,<=,=,!=,>=,>,). This compare helper lets < match to <<,
        > match to >> and = match to ==.
        """

        def __eq__(self, other):
            # type: (object) -> bool
            if str.__eq__(self, other):
                return True
            elif str.__eq__(self, "<"):
                return str.__eq__("<<", other)
            elif str.__eq__(self, ">"):
                return str.__eq__(">>", other)
            elif str.__eq__(self, "="):
                return str.__eq__("==", other)
            else:
                return False

        def __ne__(self, other):
            # type: (object) -> bool
            return not self.__eq__(other)

    def __init__(self, version, dep):
        # type: (Version, apt_pkg.Dependency) -> None
        self._version = version  # apt.package.Version
        self._dep = dep  # apt_pkg.Dependency

    def __str__(self):
        # type: () -> str
        return "%s: %s" % (self.rawtype, self.rawstr)

    def __repr__(self):
        # type: () -> str
        return "<BaseDependency: name:%r relation:%r version:%r rawtype:%r>" % (
            self.name,
            self.relation,
            self.version,
            self.rawtype,
        )

    @property
    def name(self):
        # type: () -> str
        """The name of the target package."""
        return self._dep.target_pkg.name

    @property
    def relation(self):
        # type: () -> str
        """The relation (<, <=, =, !=, >=, >, '') in mathematical notation.

        The empty string will be returned in case of an unversioned dependency.
        """
        return self.__dstr(self._dep.comp_type)

    @property
    def relation_deb(self):
        # type: () -> str
        """The relation (<<, <=, =, !=, >=, >>, '') in Debian notation.

        The empty string will be returned in case of an unversioned dependency.
        For more details see the Debian Policy Manual on the syntax of
        relationship fields:
        https://www.debian.org/doc/debian-policy/ch-relationships.html#s-depsyntax  # noqa

        .. versionadded:: 1.0.0
        """
        return self._dep.comp_type_deb

    @property
    def version(self):
        # type: () -> str
        """The target version or an empty string.

        Note that the version is only an empty string in case of an unversioned
        dependency. In this case the relation is also an empty string.
        """
        return self._dep.target_ver

    @property
    def target_versions(self):
        # type: () -> List[Version]
        """A list of all Version objects which satisfy this dependency.

        .. versionadded:: 1.0.0
        """
        tvers = []
        _tvers = self._dep.all_targets()  # type: List[apt_pkg.Version]
        for _tver in _tvers:  # type: apt_pkg.Version
            _pkg = _tver.parent_pkg  # type: apt_pkg.Package
            cache = self._version.package._pcache  # apt.cache.Cache
            pkg = cache._rawpkg_to_pkg(_pkg)  # apt.package.Package
            tver = Version(pkg, _tver)  # apt.package.Version
            tvers.append(tver)
        return tvers

    @property
    def installed_target_versions(self):
        # type: () -> List[Version]
        """A list of all installed Version objects which satisfy this dep.

        .. versionadded:: 1.0.0
        """
        return [tver for tver in self.target_versions if tver.is_installed]

    @property
    def rawstr(self):
        # type: () -> str
        """String represenation of the dependency.

        Returns the string representation of the dependency as it would be
        written in the debian/control file.  The string representation does not
        include the type of the dependency.

        Example for an unversioned dependency:
          python3

        Example for a versioned dependency:
          python3 >= 3.2

        .. versionadded:: 1.0.0
        """
        if self.version:
            return "%s %s %s" % (self.name, self.relation_deb, self.version)
        else:
            return self.name

    @property
    def rawtype(self):
        # type: () -> str
        """Type of the dependency.

        This should be one of 'Breaks', 'Conflicts', 'Depends', 'Enhances',
        'PreDepends', 'Recommends', 'Replaces', 'Suggests'.

        Additional types might be added in the future.
        """
        return self._dep.dep_type_untranslated

    @property
    def pre_depend(self):
        # type: () -> bool
        """Whether this is a PreDepends."""
        return self._dep.dep_type_untranslated == "PreDepends"


class Dependency(List[BaseDependency]):
    """Represent an Or-group of dependencies.

    Attributes defined here:
        or_dependencies - The possible choices
        rawstr - String represenation of the Or-group of dependencies
        rawtype - The type of the dependencies in the Or-group
        target_version - A list of Versions which satisfy this Or-group of deps
    """

    def __init__(self, version, base_deps, rawtype):
        # type: (Version, List[BaseDependency], str) -> None
        super(Dependency, self).__init__(base_deps)
        self._version = version  # apt.package.Version
        self._rawtype = rawtype

    def __str__(self):
        # type: () -> str
        return "%s: %s" % (self.rawtype, self.rawstr)

    def __repr__(self):
        # type: () -> str
        return "<Dependency: [%s]>" % (", ".join(repr(bd) for bd in self))

    @property
    def or_dependencies(self):
        # type: () -> Dependency
        return self

    @property
    def rawstr(self):
        # type: () -> str
        """String represenation of the Or-group of dependencies.

        Returns the string representation of the Or-group of dependencies as it
        would be written in the debian/control file.  The string representation
        does not include the type of the Or-group of dependencies.

        Example:
          python2 >= 2.7 | python3

        .. versionadded:: 1.0.0
        """
        return " | ".join(bd.rawstr for bd in self)

    @property
    def rawtype(self):
        # type: () -> str
        """Type of the Or-group of dependency.

        This should be one of 'Breaks', 'Conflicts', 'Depends', 'Enhances',
        'PreDepends', 'Recommends', 'Replaces', 'Suggests'.

        Additional types might be added in the future.

        .. versionadded:: 1.0.0
        """
        return self._rawtype

    @property
    def target_versions(self):
        # type: () -> List[Version]
        """A list of all Version objects which satisfy this Or-group of deps.

        .. versionadded:: 1.0.0
        """
        tvers = []  # type: List[Version]
        for bd in self:  # apt.package.Dependency
            for tver in bd.target_versions:  # apt.package.Version
                if tver not in tvers:
                    tvers.append(tver)
        return tvers

    @property
    def installed_target_versions(self):
        # type: () -> List[Version]
        """A list of all installed Version objects which satisfy this dep.

        .. versionadded:: 1.0.0
        """
        return [tver for tver in self.target_versions if tver.is_installed]


class Origin(object):
    """The origin of a version.

    Attributes defined here:
        archive   - The archive (eg. unstable)
        component - The component (eg. main)
        label     - The Label, as set in the Release file
        origin    - The Origin, as set in the Release file
        codename  - The Codename, as set in the Release file
        site      - The hostname of the site.
        trusted   - Boolean value whether this is trustworthy.
    """

    def __init__(self, pkg, packagefile):
        # type: (Package, apt_pkg.PackageFile) -> None
        self.archive = packagefile.archive
        self.component = packagefile.component
        self.label = packagefile.label
        self.origin = packagefile.origin
        self.codename = packagefile.codename
        self.site = packagefile.site
        self.not_automatic = packagefile.not_automatic
        # check the trust
        indexfile = pkg._pcache._list.find_index(packagefile)
        if indexfile and indexfile.is_trusted:
            self.trusted = True
        else:
            self.trusted = False

    def __repr__(self):
        # type: () -> str
        return (
            "<Origin component:%r archive:%r origin:%r label:%r "
            "site:%r isTrusted:%r>"
        ) % (
            self.component,
            self.archive,
            self.origin,
            self.label,
            self.site,
            self.trusted,
        )


class Record(Mapping[Any, Any]):
    """Record in a Packages file

    Represent a record as stored in a Packages file. You can use this like
    a dictionary mapping the field names of the record to their values::

        >>> record = Record("Package: python-apt\\nVersion: 0.8.0\\n\\n")
        >>> record["Package"]
        'python-apt'
        >>> record["Version"]
        '0.8.0'

    For example, to get the tasks of a package from a cache, you could do::

        package.candidate.record["Tasks"].split()

    Of course, you can also use the :attr:`Version.tasks` property.

    """

    def __init__(self, record_str):
        # type: (str) -> None
        self._rec = apt_pkg.TagSection(record_str)

    def __hash__(self):
        # type: () -> int
        return hash(self._rec)

    def __str__(self):
        # type: () -> str
        return str(self._rec)

    def __getitem__(self, key):
        # type: (str) -> str
        return self._rec[key]

    def __contains__(self, key):
        # type: (object) -> bool
        return key in self._rec

    def __iter__(self):
        # type: () -> Iterator[str]
        return iter(self._rec.keys())

    def iteritems(self):
        # type: () -> Iterable[Tuple[object, str]]
        """An iterator over the (key, value) items of the record."""
        for key in self._rec.keys():
            yield key, self._rec[key]

    def get(self, key, default=None):
        # type: (str, object) -> object
        """Return record[key] if key in record, else *default*.

        The parameter *default* must be either a string or None.
        """
        return self._rec.get(key, default)

    def has_key(self, key):
        # type: (str) -> bool
        """deprecated form of ``key in x``."""
        return key in self._rec

    def __len__(self):
        # type: () -> int
        return len(self._rec)


class Version(object):
    """Representation of a package version.

    The Version class contains all information related to a
    specific package version.

    .. versionadded:: 0.7.9
    """

    def __init__(self, package, cand):
        # type: (Package, apt_pkg.Version) -> None
        self.package = package
        self._cand = cand
        self.package._pcache._weakversions.add(self)

    def _cmp(self, other):
        # type: (Any) -> Union[int, Any]
        """Compares against another apt.Version object or a version string.

        This method behaves like Python 2's cmp builtin and returns an integer
        according to the outcome.  The return value is negative in case of
        self < other, zero if self == other and positive if self > other.

        The comparison includes the package name and architecture if other is
        an apt.Version object.  If other isn't an apt.Version object it'll be
        assumed that other is a version string (without package name/arch).

        .. versionchanged:: 1.0.0
        """
        # Assume that other is an apt.Version object.
        try:
            self_name = self.package.fullname
            other_name = other.package.fullname
            if self_name < other_name:
                return -1
            elif self_name > other_name:
                return 1
            return apt_pkg.version_compare(self._cand.ver_str, other.version)
        except AttributeError:
            # Assume that other is a string that only contains the version.
            try:
                return apt_pkg.version_compare(self._cand.ver_str, other)
            except TypeError:
                return NotImplemented

    def __eq__(self, other):
        # type: (object) -> bool
        return self._cmp(other) == 0

    def __ge__(self, other):
        # type: (Version) -> bool
        return self._cmp(other) >= 0

    def __gt__(self, other):
        # type: (Version) -> bool
        return self._cmp(other) > 0

    def __le__(self, other):
        # type: (Version) -> bool
        return self._cmp(other) <= 0

    def __lt__(self, other):
        # type: (Version) -> bool
        return self._cmp(other) < 0

    def __ne__(self, other):
        # type: (object) -> Union[bool, Any]
        try:
            return self._cmp(other) != 0
        except TypeError:
            return NotImplemented

    def __hash__(self):
        # type: () -> int
        return self._cand.hash

    def __str__(self):
        # type: () -> str
        return "%s=%s" % (self.package.name, self.version)

    def __repr__(self):
        # type: () -> str
        return "<Version: package:%r version:%r>" % (self.package.name, self.version)

    @property
    def _records(self):
        # type: () -> apt_pkg.PackageRecords
        """Internal helper that moves the Records to the right position."""
        # If changing lookup, change fetch_binary() as well
        if not self.package._pcache._records.lookup(self._cand.file_list[0]):
            raise LookupError("Could not lookup record")

        return self.package._pcache._records

    @property
    def _translated_records(self):
        # type: () -> Optional[apt_pkg.PackageRecords]
        """Internal helper to get the translated description."""
        desc_iter = self._cand.translated_description
        if self.package._pcache._records.lookup(desc_iter.file_list.pop(0)):
            return self.package._pcache._records
        return None

    @property
    def installed_size(self):
        # type: () -> int
        """Return the size of the package when installed."""
        return self._cand.installed_size

    @property
    def homepage(self):
        # type: () -> str
        """Return the homepage for the package."""
        return self._records.homepage

    @property
    def size(self):
        # type: () -> int
        """Return the size of the package."""
        return self._cand.size

    @property
    def architecture(self):
        # type: () -> str
        """Return the architecture of the package version."""
        return self._cand.arch

    @property
    def downloadable(self):
        # type: () -> bool
        """Return whether the version of the package is downloadable."""
        return bool(self._cand.downloadable)

    @property
    def is_installed(self):
        # type: () -> bool
        """Return wether this version of the package is currently installed.

        .. versionadded:: 1.0.0
        """
        inst_ver = self.package.installed
        return inst_ver is not None and inst_ver._cand.id == self._cand.id

    @property
    def version(self):
        # type: () -> str
        """Return the version as a string."""
        return self._cand.ver_str

    @property
    def summary(self):
        # type: () -> Optional[str]
        """Return the short description (one line summary)."""
        records = self._translated_records
        return records.short_desc if records is not None else None

    @property
    def raw_description(self):
        # type: () -> str
        """return the long description (raw)."""
        return self._records.long_desc

    @property
    def section(self):
        # type: () -> str
        """Return the section of the package."""
        return self._cand.section

    @property
    def description(self):
        # type: () -> str
        """Return the formatted long description.

        Return the formatted long description according to the Debian policy
        (Chapter 5.6.13).
        See http://www.debian.org/doc/debian-policy/ch-controlfields.html
        for more information.
        """
        desc = ""
        records = self._translated_records
        dsc = records.long_desc if records is not None else None

        if not dsc:
            return _("Missing description for '%s'." "Please report.") % (
                self.package.name
            )

        try:
            if not isinstance(dsc, str):
                # Only convert where needed (i.e. Python 2.X)
                dsc = dsc.decode("utf-8")
        except UnicodeDecodeError as err:
            return _(
                "Invalid unicode in description for '%s' (%s). " "Please report."
            ) % (self.package.name, err)

        lines = iter(dsc.split("\n"))
        # Skip the first line, since its a duplication of the summary
        next(lines)
        for raw_line in lines:
            if raw_line.strip() == ".":
                # The line is just line break
                if not desc.endswith("\n"):
                    desc += "\n\n"
                continue
            if raw_line.startswith("  "):
                # The line should be displayed verbatim without word wrapping
                if not desc.endswith("\n"):
                    line = "\n%s\n" % raw_line[2:]
                else:
                    line = "%s\n" % raw_line[2:]
            elif raw_line.startswith(" "):
                # The line is part of a paragraph.
                if desc.endswith("\n") or desc == "":
                    # Skip the leading white space
                    line = raw_line[1:]
                else:
                    line = raw_line
            else:
                line = raw_line
            # Add current line to the description
            desc += line
        return desc

    @property
    def source_name(self):
        # type: () -> str
        """Return the name of the source package."""
        try:
            return self._records.source_pkg or self.package.shortname
        except IndexError:
            return self.package.shortname

    @property
    def source_version(self):
        # type: () -> str
        """Return the version of the source package."""
        try:
            return self._records.source_ver or self._cand.ver_str
        except IndexError:
            return self._cand.ver_str

    @property
    def priority(self):
        # type: () -> str
        """Return the priority of the package, as string."""
        return self._cand.priority_str

    @property
    def policy_priority(self):
        # type: () -> int
        """Return the internal policy priority as a number.
        See apt_preferences(5) for more information about what it means.
        """
        return self.package._pcache._depcache.policy.get_priority(self._cand)

    @property
    def record(self):
        # type: () -> Record
        """Return a Record() object for this version.

        Return a Record() object for this version which provides access
        to the raw attributes of the candidate version
        """
        return Record(self._records.record)

    def get_dependencies(self, *types):
        # type: (str) -> List[Dependency]
        """Return a list of Dependency objects for the given types.

        Multiple types can be specified. Possible types are:
        'Breaks', 'Conflicts', 'Depends', 'Enhances', 'PreDepends',
        'Recommends', 'Replaces', 'Suggests'

        Additional types might be added in the future.
        """
        depends_list = []
        depends = self._cand.depends_list
        for type_ in types:
            try:
                for dep_ver_list in depends[type_]:
                    base_deps = []
                    for dep_or in dep_ver_list:
                        base_deps.append(BaseDependency(self, dep_or))
                    depends_list.append(Dependency(self, base_deps, type_))
            except KeyError:
                pass
        return depends_list

    @property
    def provides(self):
        # type: () -> List[str]
        """Return a list of names that this version provides."""
        return [p[0] for p in self._cand.provides_list]

    @property
    def enhances(self):
        # type: () -> List[Dependency]
        """Return the list of enhances for the package version."""
        return self.get_dependencies("Enhances")

    @property
    def dependencies(self):
        # type: () -> List[Dependency]
        """Return the dependencies of the package version."""
        return self.get_dependencies("PreDepends", "Depends")

    @property
    def recommends(self):
        # type: () -> List[Dependency]
        """Return the recommends of the package version."""
        return self.get_dependencies("Recommends")

    @property
    def suggests(self):
        # type: () -> List[Dependency]
        """Return the suggests of the package version."""
        return self.get_dependencies("Suggests")

    @property
    def origins(self):
        # type: () -> List[Origin]
        """Return a list of origins for the package version."""
        origins = []
        for packagefile, _unused in self._cand.file_list:
            origins.append(Origin(self.package, packagefile))
        return origins

    @property
    def filename(self):
        # type: () -> str
        """Return the path to the file inside the archive.

        .. versionadded:: 0.7.10
        """
        return self._records.filename

    @property
    def md5(self):
        # type: () -> str
        """Return the md5sum of the binary.

        .. versionadded:: 0.7.10
        """
        return self._records.md5_hash

    @property
    def sha1(self):
        # type: () -> str
        """Return the sha1sum of the binary.

        .. versionadded:: 0.7.10
        """
        return self._records.sha1_hash

    @property
    def sha256(self):
        # type: () -> str
        """Return the sha256sum of the binary.

        .. versionadded:: 0.7.10
        """
        return self._records.sha256_hash

    @property
    def tasks(self):
        # type: () -> Set[str]
        """Get the tasks of the package.

        A set of the names of the tasks this package belongs to.

        .. versionadded:: 0.8.0
        """
        return set(self.record["Task"].split())

    def _uris(self):
        # type: () -> Iterator[str]
        """Return an iterator over all available urls.

        .. versionadded:: 0.7.10
        """
        for packagefile, _unused in self._cand.file_list:
            indexfile = self.package._pcache._list.find_index(packagefile)
            if indexfile:
                yield indexfile.archive_uri(self._records.filename)

    @property
    def uris(self):
        # type: () -> List[str]
        """Return a list of all available uris for the binary.

        .. versionadded:: 0.7.10
        """
        return list(self._uris())

    @property
    def uri(self):
        # type: () -> Optional[str]
        """Return a single URI for the binary.

        .. versionadded:: 0.7.10
        """
        try:
            return next(iter(self._uris()))
        except StopIteration:
            return None

    def fetch_binary(self, destdir="", progress=None, allow_unauthenticated=None):
        # type: (str, Optional[AcquireProgress], Optional[bool]) -> str
        """Fetch the binary version of the package.

        The parameter *destdir* specifies the directory where the package will
        be fetched to.

        The parameter *progress* may refer to an apt_pkg.AcquireProgress()
        object. If not specified or None, apt.progress.text.AcquireProgress()
        is used.

        The keyword-only parameter *allow_unauthenticated* specifies whether
        to allow unauthenticated downloads. If not specified, it defaults to
        the configuration option `APT::Get::AllowUnauthenticated`.

        .. versionadded:: 0.7.10
        """
        if allow_unauthenticated is None:
            allow_unauthenticated = apt_pkg.config.find_b(
                "APT::Get::" "AllowUnauthenticated", False
            )
        base = os.path.basename(self._records.filename)
        destfile = os.path.join(destdir, base)
        if _file_is_same(destfile, self.size, self._records.hashes):
            logging.debug("Ignoring already existing file: %s" % destfile)
            return os.path.abspath(destfile)

        # Verify that the index is actually trusted
        pfile, offset = self._cand.file_list[0]
        index = self.package._pcache._list.find_index(pfile)

        if not (allow_unauthenticated or (index and index.is_trusted)):
            raise UntrustedError(
                "Could not fetch %s %s source package: "
                "Source %r is not trusted"
                % (
                    self.package.name,
                    self.version,
                    getattr(index, "describe", "<unkown>"),
                )
            )
        if not self.uri:
            raise ValueError("No URI for this binary.")
        hashes = self._records.hashes
        if not (allow_unauthenticated or hashes.usable):
            raise UntrustedError(
                "The item %r could not be fetched: " "No trusted hash found." % destfile
            )
        acq = apt_pkg.Acquire(progress or apt.progress.text.AcquireProgress())
        acqfile = apt_pkg.AcquireFile(
            acq, self.uri, hashes, self.size, base, destfile=destfile
        )
        acq.run()

        if acqfile.status != acqfile.STAT_DONE:
            raise FetchError(
                "The item %r could not be fetched: %s"
                % (acqfile.destfile, acqfile.error_text)
            )

        return os.path.abspath(destfile)

    def fetch_source(
        self, destdir="", progress=None, unpack=True, allow_unauthenticated=None
    ):
        # type: (str, Optional[AcquireProgress], bool, Optional[bool]) -> str
        """Get the source code of a package.

        The parameter *destdir* specifies the directory where the source will
        be fetched to.

        The parameter *progress* may refer to an apt_pkg.AcquireProgress()
        object. If not specified or None, apt.progress.text.AcquireProgress()
        is used.

        The parameter *unpack* describes whether the source should be unpacked
        (``True``) or not (``False``). By default, it is unpacked.

        If *unpack* is ``True``, the path to the extracted directory is
        returned. Otherwise, the path to the .dsc file is returned.

        The keyword-only parameter *allow_unauthenticated* specifies whether
        to allow unauthenticated downloads. If not specified, it defaults to
        the configuration option `APT::Get::AllowUnauthenticated`.
        """
        if allow_unauthenticated is None:
            allow_unauthenticated = apt_pkg.config.find_b(
                "APT::Get::" "AllowUnauthenticated", False
            )

        src = apt_pkg.SourceRecords()
        acq = apt_pkg.Acquire(progress or apt.progress.text.AcquireProgress())

        dsc = None
        record = self._records
        source_name = record.source_pkg or self.package.shortname
        source_version = record.source_ver or self._cand.ver_str
        source_lookup = src.lookup(source_name)

        while source_lookup and source_version != src.version:
            source_lookup = src.lookup(source_name)
        if not source_lookup:
            raise ValueError("No source for %r" % self)
        files = list()

        if not (allow_unauthenticated or src.index.is_trusted):
            raise UntrustedError(
                "Could not fetch %s %s source package: "
                "Source %r is not trusted"
                % (self.package.name, self.version, src.index.describe)
            )
        for fil in src.files:
            base = os.path.basename(fil.path)
            destfile = os.path.join(destdir, base)
            if fil.type == "dsc":
                dsc = destfile
            if _file_is_same(destfile, fil.size, fil.hashes):
                logging.debug("Ignoring already existing file: %s" % destfile)
                continue

            if not (allow_unauthenticated or fil.hashes.usable):
                raise UntrustedError(
                    "The item %r could not be fetched: "
                    "No trusted hash found." % destfile
                )
            files.append(
                apt_pkg.AcquireFile(
                    acq,
                    src.index.archive_uri(fil.path),
                    fil.hashes,
                    fil.size,
                    base,
                    destfile=destfile,
                )
            )
        acq.run()

        if dsc is None:
            raise ValueError("No source for %r" % self)

        for item in acq.items:
            if item.status != item.STAT_DONE:
                raise FetchError(
                    "The item %r could not be fetched: %s"
                    % (item.destfile, item.error_text)
                )

        if unpack:
            outdir = src.package + "-" + apt_pkg.upstream_version(src.version)
            outdir = os.path.join(destdir, outdir)
            subprocess.check_call(["dpkg-source", "-x", dsc, outdir])
            return os.path.abspath(outdir)
        else:
            return os.path.abspath(dsc)


class VersionList(Sequence[Version]):
    """Provide a mapping & sequence interface to all versions of a package.

    This class can be used like a dictionary, where version strings are the
    keys. It can also be used as a sequence, where integers are the keys.

    You can also convert this to a dictionary or a list, using the usual way
    of dict(version_list) or list(version_list). This is useful if you need
    to access the version objects multiple times, because they do not have to
    be recreated this way.

    Examples ('package.versions' being a version list):
        '0.7.92' in package.versions # Check whether 0.7.92 is a valid version.
        package.versions[0] # Return first version or raise IndexError
        package.versions[0:2] # Return a new VersionList for objects 0-2
        package.versions['0.7.92'] # Return version 0.7.92 or raise KeyError
        package.versions.keys() # All keys, as strings.
        max(package.versions)
    """

    def __init__(self, package, slice_=None):
        # type: (Package, Optional[slice]) -> None
        self._package = package  # apt.package.Package()
        self._versions = package._pkg.version_list  # [apt_pkg.Version(), ...]
        if slice_:
            self._versions = self._versions[slice_]

    def __getitem__(self, item):
        # type: (Union[int, slice, str]) -> Any
        # FIXME: Should not be returning Any, should have overloads; but
        # pyflakes complains
        if isinstance(item, slice):
            return self.__class__(self._package, item)
        try:
            # Sequence interface, item is an integer
            return Version(self._package, self._versions[item])  # type: ignore
        except TypeError:
            # Dictionary interface item is a string.
            for ver in self._versions:
                if ver.ver_str == item:
                    return Version(self._package, ver)
        raise KeyError("Version: %r not found." % (item))

    def __str__(self):
        # type: () -> str
        return "[%s]" % (", ".join(str(ver) for ver in self))

    def __repr__(self):
        # type: () -> str
        return "<VersionList: %r>" % self.keys()

    def __iter__(self):
        # type: () -> Iterator[Version]
        """Return an iterator over all value objects."""
        return (Version(self._package, ver) for ver in self._versions)

    def __contains__(self, item):
        # type: (object) -> bool
        if isinstance(item, Version):  # Sequence interface
            item = item.version
        # Dictionary interface.
        for ver in self._versions:
            if ver.ver_str == item:
                return True
        return False

    def __eq__(self, other):
        # type: (Any) -> bool
        return list(self) == list(other)

    def __len__(self):
        # type: () -> int
        return len(self._versions)

    # Mapping interface

    def keys(self):
        # type: () -> List[str]
        """Return a list of all versions, as strings."""
        return [ver.ver_str for ver in self._versions]

    def get(self, key, default=None):
        # type: (str, Optional[Version]) -> Optional[Version]
        """Return the key or the default."""
        try:
            return self[key]  # type: ignore  # FIXME: should be deterined automatically # noqa
        except LookupError:
            return default


class Package(object):
    """Representation of a package in a cache.

    This class provides methods and properties for working with a package. It
    lets you mark the package for installation, check if it is installed, and
    much more.
    """

    def __init__(self, pcache, pkgiter):
        # type: (apt.Cache, apt_pkg.Package) -> None
        """Init the Package object"""
        self._pkg = pkgiter
        self._pcache = pcache  # python cache in cache.py
        self._changelog = ""  # Cached changelog

    def __str__(self):
        # type: () -> str
        return self.name

    def __repr__(self):
        # type: () -> str
        return "<Package: name:%r architecture=%r id:%r>" % (
            self._pkg.name,
            self._pkg.architecture,
            self._pkg.id,
        )

    def __lt__(self, other):
        # type: (Package) -> bool
        return self.name < other.name

    @property
    def candidate(self):
        # type: () -> Optional[Version]
        """Return the candidate version of the package.

        This property is writeable to allow you to set the candidate version
        of the package. Just assign a Version() object, and it will be set as
        the candidate version.
        """
        cand = self._pcache._depcache.get_candidate_ver(self._pkg)
        if cand is not None:
            return Version(self, cand)
        return None

    @candidate.setter
    def candidate(self, version):
        # type: (Version) -> None
        """Set the candidate version of the package."""
        self._pcache.cache_pre_change()
        self._pcache._depcache.set_candidate_ver(self._pkg, version._cand)
        self._pcache.cache_post_change()

    @property
    def installed(self):
        # type: () -> Optional[Version]
        """Return the currently installed version of the package.

        .. versionadded:: 0.7.9
        """
        if self._pkg.current_ver is not None:
            return Version(self, self._pkg.current_ver)
        return None

    @property
    def name(self):
        # type: () -> str
        """Return the name of the package, possibly including architecture.

        If the package is not part of the system's preferred architecture,
        return the same as :attr:`fullname`, otherwise return the same
        as :attr:`shortname`

        .. versionchanged:: 0.7.100.3

        As part of multi-arch, this field now may include architecture
        information.
        """
        return self._pkg.get_fullname(True)

    @property
    def fullname(self):
        # type: () -> str
        """Return the name of the package, including architecture.

        Note that as for :meth:`architecture`, this returns the
        native architecture for Architecture: all packages.

        .. versionadded:: 0.7.100.3"""
        return self._pkg.get_fullname(False)

    @property
    def shortname(self):
        # type: () -> str
        """Return the name of the package, without architecture.

        .. versionadded:: 0.7.100.3"""
        return self._pkg.name

    @property
    def id(self):
        # type: () -> int
        """Return a uniq ID for the package.

        This can be used eg. to store additional information about the pkg."""
        return self._pkg.id

    @property
    def essential(self):
        # type: () -> bool
        """Return True if the package is an essential part of the system."""
        return self._pkg.essential

    def architecture(self):
        # type: () -> str
        """Return the Architecture of the package.

        Note that for Architecture: all packages, this returns the
        native architecture, as they are internally treated like native
        packages. To get the concrete architecture, look at the
        :attr:`Version.architecture` attribute.

        .. versionchanged:: 0.7.100.3
            This is now the package's architecture in the multi-arch sense,
            previously it was the architecture of the candidate version
            and deprecated.
        """
        return self._pkg.architecture

    # depcache states

    @property
    def marked_install(self):
        # type: () -> bool
        """Return ``True`` if the package is marked for install."""
        return self._pcache._depcache.marked_install(self._pkg)

    @property
    def marked_upgrade(self):
        # type: () -> bool
        """Return ``True`` if the package is marked for upgrade."""
        return self._pcache._depcache.marked_upgrade(self._pkg)

    @property
    def marked_delete(self):
        # type: () -> bool
        """Return ``True`` if the package is marked for delete."""
        return self._pcache._depcache.marked_delete(self._pkg)

    @property
    def marked_keep(self):
        # type: () -> bool
        """Return ``True`` if the package is marked for keep."""
        return self._pcache._depcache.marked_keep(self._pkg)

    @property
    def marked_downgrade(self):
        # type: () -> bool
        """Package is marked for downgrade"""
        return self._pcache._depcache.marked_downgrade(self._pkg)

    @property
    def marked_reinstall(self):
        # type: () -> bool
        """Return ``True`` if the package is marked for reinstall."""
        return self._pcache._depcache.marked_reinstall(self._pkg)

    @property
    def is_installed(self):
        # type: () -> bool
        """Return ``True`` if the package is installed."""
        return self._pkg.current_ver is not None

    @property
    def is_upgradable(self):
        # type: () -> bool
        """Return ``True`` if the package is upgradable."""
        return self.is_installed and self._pcache._depcache.is_upgradable(self._pkg)

    @property
    def is_auto_removable(self):
        # type: () -> bool
        """Return ``True`` if the package is no longer required.

        If the package has been installed automatically as a dependency of
        another package, and if no packages depend on it anymore, the package
        is no longer required.
        """
        return (
            self.is_installed or self.marked_install
        ) and self._pcache._depcache.is_garbage(self._pkg)

    @property
    def is_auto_installed(self):
        # type: () -> bool
        """Return whether the package is marked as automatically installed."""
        return self._pcache._depcache.is_auto_installed(self._pkg)

    # sizes

    @property
    def installed_files(self):
        # type: () -> List[str]
        """Return a list of files installed by the package.

        Return a list of unicode names of the files which have
        been installed by this package
        """
        for name in self.name, self.fullname:
            path = "/var/lib/dpkg/info/%s.list" % name
            try:
                with open(path, "rb") as file_list:
                    return file_list.read().decode("utf-8").split("\n")
            except EnvironmentError:
                continue

        return []

    def get_changelog(self, uri=None, cancel_lock=None):
        # type: (Optional[str], Optional[threading.Event]) -> str
        """
        Download the changelog of the package and return it as unicode
        string.

        The parameter *uri* refers to the uri of the changelog file. It may
        contain multiple named variables which will be substitued. These
        variables are (src_section, prefix, src_pkg, src_ver). An example is
        the Ubuntu changelog::

            "http://changelogs.ubuntu.com/changelogs/pool" \\
                "/%(src_section)s/%(prefix)s/%(src_pkg)s" \\
                "/%(src_pkg)s_%(src_ver)s/changelog"

        The parameter *cancel_lock* refers to an instance of threading.Event,
        which if set, prevents the download.
        """
        # Return a cached changelog if available
        if self._changelog != "":
            return self._changelog

        if not self.candidate:
            return _("The list of changes is not available")

        if uri is None:
            if self.candidate.origins[0].origin == "Debian":
                uri = (
                    "http://packages.debian.org/changelogs/pool"
                    "/%(src_section)s/%(prefix)s/%(src_pkg)s"
                    "/%(src_pkg)s_%(src_ver)s/changelog"
                )
            elif self.candidate.origins[0].origin == "Ubuntu":
                uri = (
                    "http://changelogs.ubuntu.com/changelogs/pool"
                    "/%(src_section)s/%(prefix)s/%(src_pkg)s"
                    "/%(src_pkg)s_%(src_ver)s/changelog"
                )
            else:
                res = _("The list of changes is not available")
                if isinstance(res, str):
                    return res
                else:
                    return res.decode("utf-8")

        # get the src package name
        src_pkg = self.candidate.source_name

        # assume "main" section
        src_section = "main"
        # use the section of the candidate as a starting point
        section = self.candidate.section

        # get the source version
        src_ver = self.candidate.source_version

        try:
            # try to get the source version of the pkg, this differs
            # for some (e.g. libnspr4 on ubuntu)
            # this feature only works if the correct deb-src are in the
            # sources.list otherwise we fall back to the binary version number
            src_records = apt_pkg.SourceRecords()
        except SystemError:
            pass
        else:
            while src_records.lookup(src_pkg):
                if not src_records.version:
                    continue
                if self.candidate.source_version == src_records.version:
                    # Direct match, use it and do not do more lookups.
                    src_ver = src_records.version
                    section = src_records.section
                    break
                if apt_pkg.version_compare(src_records.version, src_ver) > 0:
                    # The version is higher, it seems to match.
                    src_ver = src_records.version
                    section = src_records.section

        section_split = section.split("/", 1)
        if len(section_split) > 1:
            src_section = section_split[0]
        del section_split

        # lib is handled special
        prefix = src_pkg[0]
        if src_pkg.startswith("lib"):
            prefix = "lib" + src_pkg[3]

        # stip epoch
        src_ver_split = src_ver.split(":", 1)
        if len(src_ver_split) > 1:
            src_ver = "".join(src_ver_split[1:])
        del src_ver_split

        uri = uri % {
            "src_section": src_section,
            "prefix": prefix,
            "src_pkg": src_pkg,
            "src_ver": src_ver,
        }

        timeout = socket.getdefaulttimeout()

        # FIXME: when python2.4 vanishes from the archive,
        #        merge this into a single try..finally block (pep 341)
        try:
            try:
                # Set a timeout for the changelog download
                socket.setdefaulttimeout(2)

                # Check if the download was canceled
                if cancel_lock and cancel_lock.is_set():
                    return ""
                # FIXME: python3.2: Should be closed manually
                changelog_file = urlopen(uri)
                # do only get the lines that are new
                changelog = ""
                regexp = "^%s \\((.*)\\)(.*)$" % (re.escape(src_pkg))
                while True:
                    # Check if the download was canceled
                    if cancel_lock and cancel_lock.is_set():
                        return ""
                    # Read changelog line by line
                    line_raw = changelog_file.readline()
                    if not line_raw:
                        break
                    # The changelog is encoded in utf-8, but since there isn't
                    # any http header, urllib2 seems to treat it as ascii
                    line = line_raw.decode("utf-8")

                    # print line.encode('utf-8')
                    match = re.match(regexp, line)
                    if match:
                        # strip epoch from installed version
                        # and from changelog too
                        installed = getattr(self.installed, "version", None)
                        if installed and ":" in installed:
                            installed = installed.split(":", 1)[1]
                        changelog_ver = match.group(1)
                        if changelog_ver and ":" in changelog_ver:
                            changelog_ver = changelog_ver.split(":", 1)[1]

                        if (
                            installed
                            and apt_pkg.version_compare(changelog_ver, installed) <= 0
                        ):
                            break
                    # EOF (shouldn't really happen)
                    changelog += line

                # Print an error if we failed to extract a changelog
                if len(changelog) == 0:
                    changelog = _("The list of changes is not available")
                    if not isinstance(changelog, str):
                        changelog = changelog.decode("utf-8")
                self._changelog = changelog

            except HTTPError:
                if self.candidate.origins[0].origin == "Ubuntu":
                    res = _(
                        "The list of changes is not available yet.\n\n"
                        "Please use "
                        "http://launchpad.net/ubuntu/+source/%s/"
                        "%s/+changelog\n"
                        "until the changes become available or try again "
                        "later."
                    ) % (src_pkg, src_ver)
                else:
                    res = _("The list of changes is not available")
                if isinstance(res, str):
                    return res
                else:
                    return res.decode("utf-8")
            except (IOError, BadStatusLine):
                res = _(
                    "Failed to download the list of changes. \nPlease "
                    "check your Internet connection."
                )
                if isinstance(res, str):
                    return res
                else:
                    return res.decode("utf-8")
        finally:
            socket.setdefaulttimeout(timeout)
        return self._changelog

    @property
    def versions(self):
        # type: () -> VersionList
        """Return a VersionList() object for all available versions.

        .. versionadded:: 0.7.9
        """
        return VersionList(self)

    @property
    def is_inst_broken(self):
        # type: () -> bool
        """Return True if the to-be-installed package is broken."""
        return self._pcache._depcache.is_inst_broken(self._pkg)

    @property
    def is_now_broken(self):
        # type: () -> bool
        """Return True if the installed package is broken."""
        return self._pcache._depcache.is_now_broken(self._pkg)

    @property
    def has_config_files(self):
        # type: () -> bool
        """Checks whether the package is is the config-files state."""
        return self._pkg.current_state == apt_pkg.CURSTATE_CONFIG_FILES

    # depcache actions

    def mark_keep(self):
        # type: () -> None
        """Mark a package for keep."""
        self._pcache.cache_pre_change()
        self._pcache._depcache.mark_keep(self._pkg)
        self._pcache.cache_post_change()

    def mark_delete(self, auto_fix=True, purge=False):
        # type: (bool, bool) -> None
        """Mark a package for deletion.

        If *auto_fix* is ``True``, the resolver will be run, trying to fix
        broken packages.  This is the default.

        If *purge* is ``True``, remove the configuration files of the package
        as well.  The default is to keep the configuration.
        """
        self._pcache.cache_pre_change()
        self._pcache._depcache.mark_delete(self._pkg, purge)
        # try to fix broken stuffsta
        if auto_fix and self._pcache._depcache.broken_count > 0:
            fix = apt_pkg.ProblemResolver(self._pcache._depcache)
            fix.clear(self._pkg)
            fix.protect(self._pkg)
            fix.remove(self._pkg)
            fix.resolve()
        self._pcache.cache_post_change()

    def mark_install(self, auto_fix=True, auto_inst=True, from_user=True):
        # type: (bool, bool, bool) -> None
        """Mark a package for install.

        If *autoFix* is ``True``, the resolver will be run, trying to fix
        broken packages.  This is the default.

        If *autoInst* is ``True``, the dependencies of the packages will be
        installed automatically.  This is the default.

        If *fromUser* is ``True``, this package will not be marked as
        automatically installed. This is the default. Set it to False if you
        want to be able to automatically remove the package at a later stage
        when no other package depends on it.
        """
        self._pcache.cache_pre_change()
        self._pcache._depcache.mark_install(self._pkg, auto_inst, from_user)
        # try to fix broken stuff
        if auto_fix and self._pcache._depcache.broken_count > 0:
            fixer = apt_pkg.ProblemResolver(self._pcache._depcache)
            fixer.clear(self._pkg)
            fixer.protect(self._pkg)
            fixer.resolve(True)
        self._pcache.cache_post_change()

    def mark_upgrade(self, from_user=True):
        # type: (bool) -> None
        """Mark a package for upgrade."""
        if self.is_upgradable:
            auto = self.is_auto_installed
            self.mark_install(from_user=from_user)
            self.mark_auto(auto)
        else:
            # FIXME: we may want to throw a exception here
            sys.stderr.write(
                ("MarkUpgrade() called on a non-upgradeable pkg: " "'%s'\n")
                % self._pkg.name
            )

    def mark_auto(self, auto=True):
        # type: (bool) -> None
        """Mark a package as automatically installed.

        Call this function to mark a package as automatically installed. If the
        optional parameter *auto* is set to ``False``, the package will not be
        marked as automatically installed anymore. The default is ``True``.
        """
        self._pcache._depcache.mark_auto(self._pkg, auto)

    def commit(self, fprogress, iprogress):
        # type: (AcquireProgress, InstallProgress) -> None
        """Commit the changes.

        The parameter *fprogress* refers to a apt_pkg.AcquireProgress() object,
        like apt.progress.text.AcquireProgress().

        The parameter *iprogress* refers to an InstallProgress() object, as
        found in apt.progress.base.
        """
        self._pcache._depcache.commit(fprogress, iprogress)


@no_type_check
def _test():
    """Self-test."""
    print("Self-test for the Package modul")
    import random

    apt_pkg.init()
    progress = apt.progress.text.OpProgress()
    cache = apt.Cache(progress)
    pkg = cache["apt-utils"]
    print("Name: %s " % pkg.name)
    print("ID: %s " % pkg.id)
    print("Priority (Candidate): %s " % pkg.candidate.priority)
    print("Priority (Installed): %s " % pkg.installed.priority)
    print("Installed: %s " % pkg.installed.version)
    print("Candidate: %s " % pkg.candidate.version)
    print("CandidateDownloadable: %s" % pkg.candidate.downloadable)
    print("CandidateOrigins: %s" % pkg.candidate.origins)
    print("SourcePkg: %s " % pkg.candidate.source_name)
    print("Section: %s " % pkg.section)
    print("Summary: %s" % pkg.candidate.summary)
    print("Description (formatted) :\n%s" % pkg.candidate.description)
    print("Description (unformatted):\n%s" % pkg.candidate.raw_description)
    print("InstalledSize: %s " % pkg.candidate.installed_size)
    print("PackageSize: %s " % pkg.candidate.size)
    print("Dependencies: %s" % pkg.installed.dependencies)
    print("Recommends: %s" % pkg.installed.recommends)
    for dep in pkg.candidate.dependencies:
        print(
            ",".join(
                "%s (%s) (%s) (%s)" % (o.name, o.version, o.relation, o.pre_depend)
                for o in dep.or_dependencies
            )
        )
    print("arch: %s" % pkg.candidate.architecture)
    print("homepage: %s" % pkg.candidate.homepage)
    print("rec: ", pkg.candidate.record)

    print(cache["2vcard"].get_changelog())
    for i in True, False:
        print("Running install on random upgradable pkgs with AutoFix: ", i)
        for pkg in cache:
            if pkg.is_upgradable:
                if random.randint(0, 1) == 1:
                    pkg.mark_install(i)
        print("Broken: %s " % cache._depcache.broken_count)
        print("InstCount: %s " % cache._depcache.inst_count)

    print()
    # get a new cache
    for i in True, False:
        print("Randomly remove some packages with AutoFix: %s" % i)
        cache = apt.Cache(progress)
        for name in cache.keys():
            if random.randint(0, 1) == 1:
                try:
                    cache[name].mark_delete(i)
                except SystemError:
                    print("Error trying to remove: %s " % name)
        print("Broken: %s " % cache._depcache.broken_count)
        print("DelCount: %s " % cache._depcache.del_count)


# self-test
if __name__ == "__main__":
    _test()



#!/usr/bin/python3

import apt
import sys

pkgs = set()
cache = apt.Cache()
for pkg in cache:
    candver = cache._depcache.get_candidate_ver(pkg._pkg)
    if candver is None:
        continue
    dependslist = candver.depends_list
    for dep in list(dependslist.keys()):
        # get the list of each dependency object
        for depVerList in dependslist[dep]:
            for z in depVerList:
                # get all TargetVersions of
                # the dependency object
                for tpkg in z.all_targets():
                    if sys.argv[1] == tpkg.parent_pkg.name:
                        pkgs.add(pkg.name)

main = set()
universe = set()
for pkg in pkgs:
    cand = cache[pkg].candidate
    if "universe" in cand.section:
        universe.add(cand.source_name)
    else:
        main.add(cand.source_name)

print("main:")
print("\n".join(sorted(main)))
print()

print("universe:")
print("\n".join(sorted(universe)))



#!/usr/bin/python3
#
# Copyright (C) 2011 Julian Andres Klode <jak@debian.org>
#
# Copying and distribution of this file, with or without modification,
# are permitted in any medium without royalty provided the copyright
# notice and this notice are preserved.
"""Unit tests for verifying the correctness of apt_pkg.Configuration"""
import unittest
import apt_pkg

import testcommon


class TestConfiguration(testcommon.TestCase):
    """Test various configuration things"""

    def test_lp707416(self):
        """configuration: Test empty arguments (LP: #707416)"""
        self.assertRaises(ValueError, apt_pkg.parse_commandline, apt_pkg.config, [], [])
        self.assertRaises(
            SystemError,
            apt_pkg.parse_commandline,
            apt_pkg.config,
            [],
            ["cmd", "--arg0"],
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""Regression test for LP: #981896, LP: #659438"""
# Copyright (C) 2012 Sebastian Heinlein <devel@glatzor.de>
#
# Licensed under the GNU General Public License Version 2
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# Licensed under the GNU General Public License Version 2

__author__ = "Sebastian Heinlein <devel@glatzor.de>"

import os
import shutil
import tempfile
import unittest

import apt_pkg
import apt

import testcommon


class RegressionTestCase(testcommon.TestCase):

    """Test suite for LP: #981896, LP: #659438
    'Cannot locate a file for package X'
    """

    def setUp(self):
        testcommon.TestCase.setUp(self)
        apt_pkg.config.clear("APT::Update::Post-Invoke")
        apt_pkg.config.clear("APT::Update::Post-Invoke-Success")
        self.chroot_path = chroot_path = tempfile.mkdtemp()
        # Create a damaged status file
        self.cache = apt.cache.Cache(rootdir=chroot_path)
        with open(apt_pkg.config.find_file("Dir::State::status"), "a") as status:
            status.write(
                """Package: abrowser
Status: install reinstreq half-installed
Priority: optional
Section: admin
Version: 3.6.9+build1+nobinonly-0ubuntu1
Architecture: all"""
            )
        sources_list_path = apt_pkg.config.find_file("Dir::Etc::sourcelist")
        repo_path = os.path.abspath("./data/test-repo")
        with open(sources_list_path, "w") as sources_list:
            sources_list.write("deb [allow-insecure=yes] copy:%s /\n" % repo_path)
        # os.makedirs(os.path.join(chroot_path, "etc/apt/sources.list.d/"))
        self.cache.update(sources_list=sources_list_path)
        self.cache.open()

    def tearDown(self):
        # this resets the rootdir apt_pkg.config to ensure it does not
        # "pollute" the later tests
        apt.cache.Cache(rootdir="/")
        shutil.rmtree(self.chroot_path)

    def test_survive_reqreinst(self):
        """Test that we survive a package in require reinstallation state"""
        # this should be 82324L but python3.2 gets unhappy about the "L"
        self.assertEqual(self.cache.required_download, 82324)


if __name__ == "__main__":
    unittest.main()

# vim: ts=4 et sts=4
